/* ************************************************************************
 *   File: act.informative.c                             Part of CircleMUD *
 *  Usage: Player-level commands of an informative nature                  *
 *                                                                         *
 *  All rights reserved.  See license.doc for complete information.        *
 *                                                                         *
 *  Copyright (C) 1993, 94 by the Trustees of the Johns Hopkins University *
 *  CircleMUD is based on DikuMUD, Copyright (C) 1990, 1991.               *
 ************************************************************************ */

#include "info.h"

#include "comm.h"
#include "constants.h"
#include "db.h"
#include "handler.h"
#include "screen.h"

/* extern variables */
extern char *help;
extern int top_of_helpt;
extern struct help_index_element *help_table;
extern struct time_info_data time_info;
extern struct zone_data *zone_table;

extern char *credits;
extern char *news;
extern char *info;
extern char *motd;
extern char *imotd;
extern char *wizlist;
extern char *immlist;
extern char *policies;
extern char *handbook;

/* extern functions */
ACMD(do_action);
ACMD(do_insult);
void look_at_sky(struct char_data *ch);

/* local globals */
int *cmd_sort_info;


/*
 * Given the argument "look at <target>", figure out what object or char
 * matches the target.  First, see if there is another char in the room
 * with the name.  Then check local objs for exdescs.
 *
 * Thanks to Angus Mezick <angus@EDGIL.CCMAIL.COMPUSERVE.COM> for the
 * suggested fix to this problem.
 */
void look_at_target(struct char_data *ch, char *arg)
{
	int bits, found = FALSE, j, fnum, i = 0;
	struct char_data *found_char = NULL;
	struct obj_data *obj, *found_obj = NULL;
	char *desc;

	if (!ch->desc)
		return;

	if (!*arg) {
		send_to_char(ch, "Look at what?\r\n");
		return;
	}

	bits = generic_find(arg, FIND_OBJ_INV | FIND_OBJ_ROOM | FIND_OBJ_EQUIP |
							  FIND_CHAR_ROOM, ch, &found_char, &found_obj);

	/* Is the target a character? */
	if (found_char != NULL) {
		look_at_char(found_char, ch);
		if (ch != found_char) {
			if (CAN_SEE(found_char, ch))
				act("$n looks at you.", TRUE, ch, 0, found_char, TO_VICT);
			act("$n looks at $N.", TRUE, ch, 0, found_char, TO_NOTVICT);
		}
		return;
	}

	/* Strip off "number." from 2.foo and friends. */
	if (!(fnum = get_number(&arg))) {
		send_to_char(ch, "Look at what?\r\n");
		return;
	}

	/* Does the argument match an extra desc in the room? */
	if ((desc = find_exdesc(arg, world[IN_ROOM(ch)].ex_description)) != NULL && ++i == fnum) {
		page_string(ch->desc, desc, FALSE);
		return;
	}

	/* Does the argument match an extra desc in the char's equipment? */
	for (j = 0; j < NUM_WEARS && !found; j++)
		if (GET_EQ(ch, j) && CAN_SEE_OBJ(ch, GET_EQ(ch, j)))
			if ((desc = find_exdesc(arg, GET_EQ(ch, j)->ex_description)) != NULL && ++i == fnum) {
				send_to_char(ch, "%s", desc);
				found = TRUE;
			}

	/* Does the argument match an extra desc in the char's inventory? */
	for (obj = ch->carrying; obj && !found; obj = obj->next_content) {
		if (CAN_SEE_OBJ(ch, obj))
			if ((desc = find_exdesc(arg, obj->ex_description)) != NULL && ++i == fnum) {
				send_to_char(ch, "%s", desc);
				found = TRUE;
			}
	}

	/* Does the argument match an extra desc of an object in the room? */
	for (obj = world[IN_ROOM(ch)].contents; obj && !found; obj = obj->next_content)
		if (CAN_SEE_OBJ(ch, obj))
			if ((desc = find_exdesc(arg, obj->ex_description)) != NULL && ++i == fnum) {
				send_to_char(ch, "%s", desc);
				found = TRUE;
			}

	/* If an object was found back in generic_find */
	if (bits) {
		if (!found)
			show_obj_to_char(found_obj, ch, SHOW_OBJ_ACTION);
		else {
			show_obj_modifiers(found_obj, ch);
			send_to_char(ch, "\r\n");
		}
	} else if (!found)
		send_to_char(ch, "You do not see that here.\r\n");
}


ACMD(do_look)
{
	int look_type;

	if (!ch->desc)
		return;

	if (GET_POS(ch) < POS_SLEEPING)
		send_to_char(ch, "You can't see anything but stars!\r\n");
	else if (AFF_FLAGGED(ch, AFF_BLIND))
		send_to_char(ch, "You can't see a damned thing, you're blind!\r\n");
	else if (ROOM_AFFECTED(IN_ROOM(ch), RAFF_FOG) && GET_LEVEL(ch) < LVL_IMMORT)
	  send_to_char(ch, "Your view is obscured by a mysterious fog.\r\n");
	else if (IS_DARK(IN_ROOM(ch)) && !CAN_SEE_IN_DARK(ch)) {
		send_to_char(ch, "It is pitch black...\r\n");
		list_char_to_char(world[IN_ROOM(ch)].people, ch);	/* glowing red eyes */
	} else {
		char arg[MAX_INPUT_LENGTH], arg2[MAX_INPUT_LENGTH];

		half_chop(argument, arg, arg2);

		if (subcmd == SCMD_READ) {
			if (!*arg)
				send_to_char(ch, "Read what?\r\n");
			else
				look_at_target(ch, arg);
			return;
		}
		if (!*arg)			/* "look" alone, without an argument at all */
			look_at_room(ch, 1);
		else if(is_abbrev(arg, "sky")) {
			if (!OUTSIDE(ch) && GET_LEVEL(ch) < LVL_GOD)
				send_to_char(ch, "You can't see the sky indoors.\n\r");
			else 
				look_at_sky(ch);
		}
		else if (is_abbrev(arg, "in"))
			look_in_obj(ch, arg2);
		/* did the char type 'look <direction>?' */
		else if ((look_type = search_block(arg, dirs, FALSE)) >= 0)
			look_in_direction(ch, look_type);
		else if (is_abbrev(arg, "at"))
			look_at_target(ch, arg2);
		else
			look_at_target(ch, arg);
	}
}



ACMD(do_time)
{
	const char *suf;
	int weekday, day;

	/* day in [1..35] */
	day = time_info.day + 1;

	/* 35 days in a month, 7 days a week */
	weekday = ((35 * time_info.month) + day) % 7;

	send_to_char(ch, "It is %d o'clock %s, on %s.\r\n",
					 (time_info.hours % 12 == 0) ? 12 : (time_info.hours % 12),
					 time_info.hours >= 12 ? "pm" : "am", weekdays[weekday]);

	/*
	 * Peter Ajamian <peter@PAJAMIAN.DHS.ORG> supplied the following as a fix
	 * for a bug introduced in the ordinal display that caused 11, 12, and 13
	 * to be incorrectly displayed as 11st, 12nd, and 13rd.  Nate Winters
	 * <wintersn@HOTMAIL.COM> had already submitted a fix, but it hard-coded a
	 * limit on ordinal display which I want to avoid.	-dak
	 */

	suf = "th";

	if (((day % 100) / 10) != 1) {
		switch (day % 10) {
			case 1:
				suf = "st";
				break;
			case 2:
				suf = "nd";
				break;
			case 3:
				suf = "rd";
				break;
		}
	}

	send_to_char(ch, "The %d%s Day of the %s, Year %d.\r\n",
					 day, suf, month_name[time_info.month], time_info.year);
}


ACMD(do_weather)
{
	if (OUTSIDE(ch)) {
		send_to_char(ch, "The sky is %s and %s.\r\n", 
						 sky_look[zone_table[IN_ZONE(ch)].sky],
						 (zone_table[IN_ZONE(ch)].change >= 0 ? 
						  "you feel a warm wind from south" :
						  "your foot tells you bad weather is due"));

		if (GET_LEVEL(ch) >= LVL_GOD)
			send_to_char(ch, "Pressure: %d (change: %d), Sky: %d (%s)\r\n",
							 zone_table[IN_ZONE(ch)].pressure, 
							 zone_table[IN_ZONE(ch)].change, 
							 zone_table[IN_ZONE(ch)].sky,
							 sky_look[zone_table[IN_ZONE(ch)].sky]);
	} else
		send_to_char(ch, "You have no feeling about the weather at all.\r\n");
}


ACMD(do_help)
{
	int chk, bot, top, mid, minlen;

	if (!ch->desc)
		return;

	skip_spaces(&argument);

	if (!*argument) {
		page_string(ch->desc, help, 0);
		return;
	}
	if (!help_table) {
		send_to_char(ch, "No help available.\r\n");
		return;
	}

	bot = 0;
	top = top_of_helpt;
	minlen = strlen(argument);

	for (;;) {
		mid = (bot + top) / 2;

		if (bot > top) {
			send_to_char(ch, "There is no help on that word.\r\n");
			mudlog(NRM, LVL_GRGOD, TRUE, 
					 "HELP: %s could not find help on %s", GET_NAME(ch), argument);
			return;
		} else if (!(chk = strn_cmp(argument, help_table[mid].keyword, minlen))) {
			/* trace backwards to find first matching entry. Thanks Jeff Fink! */
			while ((mid > 0) &&
					 (!(chk = strn_cmp(argument, help_table[mid - 1].keyword, minlen))))
				mid--;
			page_string(ch->desc, help_table[mid].entry, 0);
			return;
		} else {
			if (chk > 0)
				bot = mid + 1;
			else
				top = mid - 1;
		}
	}
}


/* Generic page_string function for displaying text */
ACMD(do_gen_ps)
{
	switch (subcmd) {
		case SCMD_CREDITS:
			page_string(ch->desc, credits, 0);
			break;
		case SCMD_NEWS:
			page_string(ch->desc, news, 0);
			break;
		case SCMD_INFO:
			page_string(ch->desc, info, 0);
			break;
		case SCMD_WIZLIST:
			page_string(ch->desc, wizlist, 0);
			break;
		case SCMD_IMMLIST:
			page_string(ch->desc, immlist, 0);
			break;
		case SCMD_HANDBOOK:
			page_string(ch->desc, handbook, 0);
			break;
		case SCMD_POLICIES:
			page_string(ch->desc, policies, 0);
			break;
		case SCMD_MOTD:
			page_string(ch->desc, motd, 0);
			break;
		case SCMD_IMOTD:
			page_string(ch->desc, imotd, 0);
			break;
		case SCMD_CLEAR:
			send_to_char(ch, "\033[H\033[J");
			break;
		case SCMD_VERSION:
			send_to_char(ch, "%s\r\n", circlemud_version);
			break;
		case SCMD_WHOAMI:
			send_to_char(ch, "%s\r\n", GET_NAME(ch));
			break;
		default:
			log("SYSERR: Unhandled case in do_gen_ps. (%d)", subcmd);
			return;
	}
}


const char *ctypes[] = {
	"off", "sparse", "normal", "complete", "\n"
};

ACMD(do_color)
{
	char arg[MAX_INPUT_LENGTH];
	int tp;

	if (IS_NPC(ch))
		return;

	one_argument(argument, arg);

	if (!*arg) {
		send_to_char(ch, "Your current color level is %s.\r\n", ctypes[COLOR_LEV(ch)]);
		return;
	}
	if (((tp = search_block(arg, ctypes, FALSE)) == -1)) {
		send_to_char(ch, "Usage: color { Off | Sparse | Normal | Complete }\r\n");
		return;
	}
	PRF_FLAGS(ch).PRF_COLOR_1 = tp & 1;
	PRF_FLAGS(ch).PRF_COLOR_2 = (tp & 2) >> 1;

	send_to_char(ch, "Your %scolor%s is now %s.\r\n", 
					 QRED, CCNRM(ch, C_OFF), ctypes[tp]);
}


ACMD(do_toggle)
{
	char buf2[4];

	if (IS_NPC(ch))
		return;

	if (GET_WIMP_LEV(ch) == 0)
		strlcpy(buf2, "OFF", sizeof(buf2));
	else
		snprintf(buf2, sizeof(buf2), "%-3.3d", GET_WIMP_LEV(ch));

	if (GET_LEVEL(ch) >= LVL_IMMORT) {
		send_to_char(ch,
						 "      No Hassle: %-3s    "
						 "      Holylight: %-3s    "
						 "     Room Flags: %-3s\r\n",
						 ONOFF(PRF_FLAGGED(ch, PRF_NOHASSLE)),
						 ONOFF(PRF_FLAGGED(ch, PRF_HOLYLIGHT)),
						 ONOFF(PRF_FLAGGED(ch, PRF_ROOMFLAGS))
						 );
	}

	send_to_char(ch,
					 "Hit Pnt Display: %-3s    "
					 "     Brief Mode: %-3s    "
					 " Summon Protect: %-3s\r\n"

					 "   Move Display: %-3s    "
					 "   Compact Mode: %-3s    "
					 "       On Quest: %-3s\r\n"

					 "   Mana Display: %-3s    "
					 "         NoTell: %-3s    "
					 "   Repeat Comm.: %-3s\r\n"

                "    Exp Display: %-3s    "
                " Auto Show Exit: %-3s    "
                "      Auto Loot: %-3s\r\n"

					 "   Gold Display: %-3s    "
                "    Auto Assist: %-3s    "
                "     Auto Split: %-3s\r\n"

                " Auto Sacrifice: %-3s\r\n"

                " Gossip Channel: %-3s    "
                "Auction Channel: %-3s    "
                "  Grats Channel: %-3s\r\n"

                "           Deaf: %-3s    "
                "     Wimp Level: %-3s\r\n"

                "    Color Level: %s\r\n",

					 ONOFF(PRF_FLAGGED(ch, PRF_DISPHP)),
					 ONOFF(PRF_FLAGGED(ch, PRF_BRIEF)),
					 ONOFF(!PRF_FLAGGED(ch, PRF_SUMMONABLE)),

					 ONOFF(PRF_FLAGGED(ch, PRF_DISPMOVE)),
					 ONOFF(PRF_FLAGGED(ch, PRF_COMPACT)),
					 YESNO(PRF_FLAGGED(ch, PRF_QUEST)),

					 ONOFF(PRF_FLAGGED(ch, PRF_DISPMANA)),
					 ONOFF(PRF_FLAGGED(ch, PRF_NOTELL)),
					 YESNO(!PRF_FLAGGED(ch, PRF_NOREPEAT)),

					 ONOFF(PRF_FLAGGED(ch, PRF_DISPEXP)),
					 ONOFF(PRF_FLAGGED(ch, PRF_AUTOEXIT)),
					 ONOFF(PRF_FLAGGED(ch, PRF_AUTOLOOT)),

					 ONOFF(PRF_FLAGGED(ch, PRF_DISPGOLD)),
					 ONOFF(PRF_FLAGGED(ch, PRF_AUTOASSIST)),
					 ONOFF(PRF_FLAGGED(ch, PRF_AUTOSPLIT)),

					 ONOFF(PRF_FLAGGED(ch, PRF_AUTOSAC)),

					 ONOFF(!PRF_FLAGGED(ch, PRF_NOGOSS)),
					 ONOFF(!PRF_FLAGGED(ch, PRF_NOAUCT)),
					 ONOFF(!PRF_FLAGGED(ch, PRF_NOGRATZ)),

					 YESNO(PRF_FLAGGED(ch, PRF_DEAF)),
					 buf2,
					 ctypes[COLOR_LEV(ch)]);
}


int sort_commands_helper(const void *a, const void *b)
{
	return strcmp(cmd_info[*(const int *)a].command, cmd_info[*(const int *)b].command);
}


void sort_commands(void)
{
	int a, num_of_cmds = 0;

	while (cmd_info[num_of_cmds].command[0] != '\n')
		num_of_cmds++;
	num_of_cmds++;	/* \n */

	CREATE(cmd_sort_info, int, num_of_cmds);

	for (a = 0; a < num_of_cmds; a++)
		cmd_sort_info[a] = a;

	/* Don't sort the RESERVED or \n entries. */
	qsort(cmd_sort_info + 1, num_of_cmds - 2, sizeof(int), sort_commands_helper);
}


ACMD(do_commands)
{
	int no, i, cmd_num;
	int wizhelp = 0, socials = 0;
	struct char_data *vict;
	char arg[MAX_INPUT_LENGTH];

	one_argument(argument, arg);

	if (*arg) {
		if (!(vict = get_char_vis(ch, arg, NULL, FIND_CHAR_WORLD)) || IS_NPC(vict)) {
			send_to_char(ch, "Who is that?\r\n");
			return;
		}
		if (GET_LEVEL(ch) < GET_LEVEL(vict)) {
			send_to_char(ch, "You can't see the commands of people above your level.\r\n");
			return;
		}
	} else
		vict = ch;

	if (subcmd == SCMD_SOCIALS)
		socials = 1;
	else if (subcmd == SCMD_WIZHELP)
		wizhelp = 1;

	send_to_char(ch, "The following %s%s are available to %s:\r\n",
					 wizhelp ? "privileged " : "",
					 socials ? "socials" : "commands",
					 vict == ch ? "you" : GET_NAME(vict));

	/* cmd_num starts at 1, not 0, to remove 'RESERVED' */
	for (no = 1, cmd_num = 1; cmd_info[cmd_sort_info[cmd_num]].command[0] != '\n'; cmd_num++) {
		i = cmd_sort_info[cmd_num];

		if (cmd_info[i].minimum_level < 0 || GET_LEVEL(vict) < cmd_info[i].minimum_level)
			continue;

		if ((cmd_info[i].minimum_level >= LVL_IMMORT) != wizhelp)
			continue;

		if (!wizhelp && socials != (cmd_info[i].command_pointer == do_action || cmd_info[i].command_pointer == do_insult))
			continue;

		send_to_char(ch, "%-11s%s", cmd_info[i].command, no++ % 7 == 0 ? "\r\n" : "");
	}

	if (no % 7 != 1)
		send_to_char(ch, "\r\n");
}


ACMD(do_index)
{
   int i, row = 0, minlen;
	size_t nlen = 0, len = 0;
	char buf[MAX_STRING_LENGTH];

   if (!ch->desc)
      return;

   skip_spaces(&argument);

   if (!*argument) {
      send_to_char(ch, "USAGE: index <letter|phrase>\r\n");
      return;
	}

	if (!help_table) {
		send_to_char(ch, "No help available.\r\n");
		return;
	}

   minlen = strlen(argument);

   for (i = 0; i < top_of_helpt; i++) {
      if (!strn_cmp(argument, help_table[i].keyword, minlen)) {
			row += 1;
			nlen = snprintf(buf + len, sizeof(buf) - len, 
								 "| %-21.21s |%s", help_table[i].keyword,
								 (((row % 3) == 0) ? "\r\n" : ""));

			if (len + nlen >= sizeof(buf) || nlen < 0)
				break;
			len += nlen;
		}
	}

   if (ch->desc) {
      page_string(ch->desc, buf, TRUE);
		send_to_char(ch, "\r\n");
	}
}

