/* ************************************************************************
 *   File: class.c                                       Part of CircleMUD *
 *  Usage: Source file for class-specific code                             *
 *                                                                         *
 *  All rights reserved.  See license.doc for complete information.        *
 *                                                                         *
 *  Copyright (C) 1993, 94 by the Trustees of the Johns Hopkins University *
 *  CircleMUD is based on DikuMUD, Copyright (C) 1990, 1991.               *
 ************************************************************************ */

/*
 * This file attempts to concentrate most of the code which must be changed
 * in order for new classes to be added.  If you're adding a new class,
 * you should go through this entire file from beginning to end and add
 * the appropriate new special cases for your new class.
 */

#include "conf.h"
#include "sysdep.h"

#include "structs.h"
#include "utils.h"

#include "constants.h"
#include "db.h"
#include "interpreter.h"
#include "spells.h"


extern int siteok_everyone;
extern int allow_stat_edit;

/* extern functions */
void racial_ability_modifiers(struct char_data *ch);
void set_height_by_race(struct char_data *ch);
void set_weight_by_race(struct char_data *ch);

/* local functions */
void snoop_check(struct char_data *ch);
int parse_class(struct char_data *ch, char arg);
bitvector_t find_class_bitvector(const char *arg);
byte saving_throws(int class_num, int type, int level);
int thaco(int class_num, int level);
void roll_real_abils(struct char_data *ch);
void do_start(struct char_data *ch);
int backstab_mult(int level);
int invalid_class(struct char_data *ch, struct obj_data *obj);
int level_exp(int chclass, int level);
const char *title_male(int chclass, int level);
const char *title_female(int chclass, int level);

/* Names first */

const char *class_abbrevs[] = {
	"Mu",
	"Cl",
	"Th",
	"Wa",
	"\n"
};


const char *pc_class_types[] = {
	"Magic User",
	"Cleric",
	"Thief",
	"Warrior",
	"\n"
};


/* The menu for choosing a class in interpreter.c: */
const char *class_menu =
"\r\n"
"Select a class:\r\n"
"  [C]leric\r\n"
"  [T]hief\r\n"
"  [W]arrior\r\n"
"  [M]agic-user\r\n";



/*
 * The code to interpret a class letter -- used in interpreter.c when a
 * new character is selecting a class and by 'set class' in act.wizard.c.
 *
 * I added race limits on classes.  Each race is restricted by one
 * class, except for humans.  Have to give humans SOME advantage :p
 *  -ejg
 */

int parse_class(struct char_data *ch, char arg)
{
	arg = LOWER(arg);

	switch (arg) {
		case 'm': /* No dwarven magic users */
			if (!ch || GET_RACE(ch) != RACE_DWARF)
				return CLASS_MAGIC_USER;

			return CLASS_UNDEFINED;

		case 'c': /* No gnome clerics */
			if (!ch || GET_RACE(ch) != RACE_GNOME)
				return CLASS_CLERIC;

			return CLASS_UNDEFINED;

		case 'w':
			return CLASS_WARRIOR;

		case 't': /* No elven thieves - it runs counter to their culture! */
			if (!ch || GET_RACE(ch) != RACE_ELF)
				return CLASS_THIEF;

			return CLASS_UNDEFINED;

		default:
			return CLASS_UNDEFINED;
	}
}

/*
 * bitvectors (i.e., powers of two) for each class, mainly for use in
 * do_who and do_users.  Add new classes at the end so that all classes
 * use sequential powers of two (1 << 0, 1 << 1, 1 << 2, 1 << 3, 1 << 4,
 * 1 << 5, etc.) up to the limit of your bitvector_t, typically 0-31.
 */
bitvector_t find_class_bitvector(const char *arg)
{
	size_t rpos, ret = 0;

	for (rpos = 0; rpos < strlen(arg); rpos++)
		ret |= (1 << parse_class(NULL, arg[rpos]));

	return (ret);
}


/*
 * These are definitions which control the guildmasters for each class.
 *
 * The first field (top line) controls the highest percentage skill level
 * a character of the class is allowed to attain in any skill.  (After
 * this level, attempts to practice will say "You are already learned in
 * this area."
 * 
 * The second line controls the maximum percent gain in learnedness a
 * character is allowed per practice -- in other words, if the random
 * die throw comes out higher than this number, the gain will only be
 * this number instead.
 *
 * The third line controls the minimu percent gain in learnedness a
 * character is allowed per practice -- in other words, if the random
 * die throw comes out below this number, the gain will be set up to
 * this number.
 * 
 * The fourth line simply sets whether the character knows 'spells'
 * or 'skills'.  This does not affect anything except the message given
 * to the character when trying to practice (i.e. "You know of the
 * following spells" vs. "You know of the following skills"
 */

#define SPELL	0
#define SKILL	1

/* #define LEARNED_LEVEL	0  % known which is considered "learned" */
/* #define MAX_PER_PRAC		1  max percent gain in skill per practice */
/* #define MIN_PER_PRAC		2  min percent gain in skill per practice */
/* #define PRAC_TYPE		3  should it say 'spell' or 'skill'?	*/

int prac_params[4][NUM_CLASSES] = {
	/* MAG	CLE	THE	WAR */
	{ 95,		95,	85,	80	},	/* learned level */
	{ 100,	100,	12,	12	},	/* max per practice */
	{ 25,		25,	0,	0	},	/* min per practice */
	{ SPELL,	SPELL,	SKILL,	SKILL	},	/* prac name */
};


/*
 * ...And the appropriate rooms for each guildmaster/guildguard; controls
 * which types of people the various guildguards let through.  i.e., the
 * first line shows that from room 3017, only MAGIC_USERS are allowed
 * to go south.
 *
 * Don't forget to visit spec_assign.c if you create any new mobiles that
 * should be a guild master or guard so they can act appropriately. If you
 * "recycle" the existing mobs that are used in other guilds for your new
 * guild, then you don't have to change that file, only here.
 */
struct guild_info_type guild_info[] = {

/* Midgaard */
	{ CLASS_MAGIC_USER,	3017,	SCMD_SOUTH	},
	{ CLASS_CLERIC,	3004,	SCMD_NORTH	},
	{ CLASS_THIEF,	3027,	SCMD_EAST	},
	{ CLASS_WARRIOR,	3021,	SCMD_EAST	},

/* Brass Dragon */
	{ -999 /* all */ ,	5065,	SCMD_WEST	},

/* this must go last -- add new guards above! */
	{ -1, NOWHERE, -1}
};


/* Order of saving throws: PARA, ROD, PETRI, BREATH, SPELL */
/* This matrix determines the base saving throw for each   */
/* class (i.e. the saving throw needed at level 1).        */
int saving_throw_base[NUM_CLASSES][5] = {
	{ 70, 55, 65, 75, 60 }, /* MU */
	{ 60, 70, 65, 80, 75 }, /* CL */
	{ 65, 70, 60, 80, 75 }, /* TH */
	{ 70, 80, 75, 85, 85 }, /* WA */
};        

/* This is the rate at which saving throws improve, *10 */
/* So, if a saving throw improves at a rate of 1.4 per  */
/* level, it is entered into this matrix as -14.        */
int saving_throw_improve[NUM_CLASSES][5] = {
	{ -14, -16, -17, -17, -17 }, /* MU */
	{ -16, -15, -15, -15, -15 }, /* CL */
	{ -10, -20, -10, -10, -20 }, /* TH */
	{ -21, -21, -21, -25, -21 }, /* WA */
};


byte saving_throws(int class_num, int type, int level)
{
	float mod = (saving_throw_improve[class_num][type] / 10.0f) * (level - 1);
  
	return (saving_throw_base[class_num][type] + mod);
}


/* THAC0 for classes and levels.  (To Hit Armor Class 0) */
int thaco(int class_num, int level)
{
	int t = 100; /* Default THAC0 */
  
	if (level > 0) {
		switch (class_num) {
			case CLASS_MAGIC_USER: t = 20 - ((level - 1) / 3);       break;
			case CLASS_CLERIC:     t = 20 - (((level - 1) / 3) * 2); break;
			case CLASS_THIEF:      t = 20 - ((level - 1) / 2);       break;
			case CLASS_WARRIOR:    t = 20 - (level - 1);             break;
			default:
				log("SYSERR: Unknown class in thac0 chart.");
		}
	}
	
	if (t < 1) /* Nobody can have a THAC0 better than 1! */
		t = 1;
    
	return (t);
}


/*
 * Roll the 6 stats for a character... each stat is made of the sum of
 * the best 3 out of 4 rolls of a 6-sided die.  Each class then decides
 * which priority will be given for the best to worst stats.
 */
void roll_real_abils(struct char_data *ch)
{
	int i, j, k, temp;
	ubyte table[6];
	ubyte rolls[4];

	if (!allow_stat_edit) {
		for (i = 0; i < 6; i++)
			table[i] = 0;

		for (i = 0; i < 6; i++) {

			for (j = 0; j < 4; j++)
				rolls[j] = rand_number(1, 6);

			temp = rolls[0] + rolls[1] + rolls[2] + rolls[3] -
				MIN(rolls[0], MIN(rolls[1], MIN(rolls[2], rolls[3])));

			for (k = 0; k < 6; k++)
				if (table[k] < temp) {
					temp ^= table[k];
					table[k] ^= temp;
					temp ^= table[k];
				}
		}

		ch->real_abils.str_add = 0;

		switch (GET_CLASS(ch)) {
			case CLASS_MAGIC_USER:
				ch->real_abils.intel = table[0];
				ch->real_abils.wis = table[1];
				ch->real_abils.dex = table[2];
				ch->real_abils.str = table[3];
				ch->real_abils.con = table[4];
				ch->real_abils.cha = table[5];
				break;
			case CLASS_CLERIC:
				ch->real_abils.wis = table[0];
				ch->real_abils.intel = table[1];
				ch->real_abils.str = table[2];
				ch->real_abils.dex = table[3];
				ch->real_abils.con = table[4];
				ch->real_abils.cha = table[5];
				break;
			case CLASS_THIEF:
				ch->real_abils.dex = table[0];
				ch->real_abils.str = table[1];
				ch->real_abils.con = table[2];
				ch->real_abils.intel = table[3];
				ch->real_abils.wis = table[4];
				ch->real_abils.cha = table[5];
				break;
			case CLASS_WARRIOR:
				ch->real_abils.str = table[0];
				ch->real_abils.dex = table[1];
				ch->real_abils.con = table[2];
				ch->real_abils.wis = table[3];
				ch->real_abils.intel = table[4];
				ch->real_abils.cha = table[5];
				if (ch->real_abils.str == 18)
					ch->real_abils.str_add = rand_number(0, 100);
				break;
		}
	}
	
	racial_ability_modifiers(ch);
	set_height_by_race(ch);
	set_weight_by_race(ch);

	ch->aff_abils = ch->real_abils;
}


/* Some initializations for characters, including initial skills */
void do_start(struct char_data *ch)
{
	GET_LEVEL(ch) = 1;
	GET_EXP(ch) = 1;

	set_title(ch, NULL);
	roll_real_abils(ch);

	GET_MAX_HIT(ch)  = 10;
	GET_MAX_MANA(ch) = 100;
	GET_MAX_MOVE(ch) = 82;

	switch (GET_CLASS(ch)) {

		case CLASS_MAGIC_USER:
			SET_SKILL(ch, SKILL_BANDAGE, 5);
			SET_SKILL(ch, SKILL_MOUNT, 5);
			SET_SKILL(ch, SKILL_RIDING, 10);
			break;

		case CLASS_CLERIC:
			SET_SKILL(ch, SKILL_BANDAGE, 25);
			SET_SKILL(ch, SKILL_MOUNT, 10);
			SET_SKILL(ch, SKILL_RIDING, 15);
			break;

		case CLASS_THIEF:
			SET_SKILL(ch, SKILL_BACKSTAB, 10);
			SET_SKILL(ch, SKILL_BANDAGE, 10);
			SET_SKILL(ch, SKILL_MOUNT, 20);
			SET_SKILL(ch, SKILL_RIDING, 25);
			SET_SKILL(ch, SKILL_HIDE, 5);
			SET_SKILL(ch, SKILL_LISTEN, 15);
			SET_SKILL(ch, SKILL_PICK_LOCK, 10);
			SET_SKILL(ch, SKILL_SAP, 10);
			SET_SKILL(ch, SKILL_SNEAK, 10);
			SET_SKILL(ch, SKILL_STEAL, 15);
			SET_SKILL(ch, SKILL_TRACK, 10);
			SET_SKILL(ch, SKILL_THROW, 5);
			break;

		case CLASS_WARRIOR:
			SET_SKILL(ch, SKILL_BANDAGE, 20);
			SET_SKILL(ch, SKILL_MOUNT, 50);
			SET_SKILL(ch, SKILL_RIDING, 50);
			SET_SKILL(ch, SKILL_THROW, 15);
			break;
	}

	advance_level(ch);
	mudlog(BRF, MAX(LVL_IMMORT, GET_INVIS_LEV(ch)), TRUE,
			 "%s advanced to level %d", GET_NAME(ch), GET_LEVEL(ch));

	GET_HIT(ch) = GET_MAX_HIT(ch);
	GET_MANA(ch) = GET_MAX_MANA(ch);
	GET_MOVE(ch) = GET_MAX_MOVE(ch);

	GET_COND(ch, THIRST) = 24;
	GET_COND(ch, FULL) = 24;
	GET_COND(ch, DRUNK) = 0;

	if (siteok_everyone)
		PLR_FLAGS(ch).PLR_SITEOK = 1;
}



/*
 * This function controls the change to maxmove, maxmana, and maxhp for
 * each class every time they gain a level.
 */
void advance_level(struct char_data *ch)
{
	int add_hp, add_mana = 0, add_move = 0, i;

	add_hp = con_app[GET_CON(ch)].hitp;

	switch (GET_CLASS(ch)) {

		case CLASS_MAGIC_USER:
			add_hp += rand_number(3, 8);
			add_mana = rand_number(GET_LEVEL(ch), (int)(1.5 * GET_LEVEL(ch)));
			add_mana = MIN(add_mana, 10);
			add_move = rand_number(0, 2);
			break;

		case CLASS_CLERIC:
			add_hp += rand_number(5, 10);
			add_mana = rand_number(GET_LEVEL(ch), (int)(1.5 * GET_LEVEL(ch)));
			add_mana = MIN(add_mana, 10);
			add_move = rand_number(0, 2);
			break;

		case CLASS_THIEF:
			add_hp += rand_number(7, 13);
			add_mana = 0;
			add_move = rand_number(1, 3);
			break;

		case CLASS_WARRIOR:
			add_hp += rand_number(10, 15);
			add_mana = 0;
			add_move = rand_number(1, 3);
			break;
	}

	ch->points.max_hit += MAX(1, add_hp);
	ch->points.max_move += MAX(1, add_move);

	if (GET_LEVEL(ch) > 1)
		ch->points.max_mana += add_mana;

	if (IS_MAGIC_USER(ch) || IS_CLERIC(ch))
		GET_PRACTICES(ch) += MAX(2, wis_app[GET_WIS(ch)].bonus);
	else
		GET_PRACTICES(ch) += MIN(2, MAX(1, wis_app[GET_WIS(ch)].bonus));

	if (GET_LEVEL(ch) >= LVL_IMMORT) {
		for (i = 0; i < 3; i++)
			GET_COND(ch, i) = (char) -1;
		PRF_FLAGS(ch).PRF_HOLYLIGHT = 1;
	}

	snoop_check(ch);
	save_char(ch);
}


/*
 * This simply calculates the backstab multiplier based on a character's
 * level.  This used to be an array, but was changed to be a function so
 * that it would be easier to add more levels to your MUD.  This doesn't
 * really create a big performance hit because it's not used very often.
 */
int backstab_mult(int level)
{
	if (level <= 0)
		return 1;	  /* level 0 */
	else if (level <= 7)
		return 2;	  /* level 1 - 7 */
	else if (level <= 13)
		return 3;	  /* level 8 - 13 */
	else if (level <= 20)
		return 4;	  /* level 14 - 20 */
	else if (level <= 28)
		return 5;	  /* level 21 - 28 */
	else if (level < LVL_IMMORT)
		return 6;	  /* all remaining mortal levels */
	else
		return 20;	  /* immortals */
}


/*
 * invalid_class is used by handler.c to determine if a piece of equipment is
 * usable by a particular class, based on the ITEM_ANTI_{class} bitvectors.
 */
int invalid_class(struct char_data *ch, struct obj_data *obj)
{
	if (GET_LEVEL(ch) >= LVL_IMMORT)
		return FALSE;
	
	if (OBJ_FLAGGED(obj, ITEM_ANTI_MAGIC_USER) && IS_MAGIC_USER(ch))
		return TRUE;

	if (OBJ_FLAGGED(obj, ITEM_ANTI_CLERIC) && IS_CLERIC(ch))
		return TRUE;

	if (OBJ_FLAGGED(obj, ITEM_ANTI_WARRIOR) && IS_WARRIOR(ch))
		return TRUE;

	if (OBJ_FLAGGED(obj, ITEM_ANTI_THIEF) && IS_THIEF(ch))
		return TRUE;

	return FALSE;
}


/*
 * SPELLS AND SKILLS.  This area defines which spells are assigned to
 * which classes, and the minimum level the character must be to use
 * the spell or skill.
 */
void init_spell_levels(void)
{
	/* MAGES */
	spell_level(SPELL_MAGIC_MISSILE, CLASS_MAGIC_USER, 1);
	spell_level(SPELL_DETECT_INVIS, CLASS_MAGIC_USER, 2);
	spell_level(SPELL_DETECT_MAGIC, CLASS_MAGIC_USER, 2);
	spell_level(SPELL_CHILL_TOUCH, CLASS_MAGIC_USER, 3);
	spell_level(SPELL_INFRAVISION, CLASS_MAGIC_USER, 3);
	spell_level(SPELL_INVISIBLE, CLASS_MAGIC_USER, 4);
	spell_level(SPELL_ARMOR, CLASS_MAGIC_USER, 4);
	spell_level(SPELL_BURNING_HANDS, CLASS_MAGIC_USER, 5);
	spell_level(SPELL_VENTRILOQUATE, CLASS_MAGIC_USER, 5);
	spell_level(SPELL_LOCATE_OBJECT, CLASS_MAGIC_USER, 6);
	spell_level(SPELL_STRENGTH, CLASS_MAGIC_USER, 6);
	spell_level(SPELL_FEAR, CLASS_MAGIC_USER, 7);
	spell_level(SPELL_SHOCKING_GRASP, CLASS_MAGIC_USER, 7);
	spell_level(SPELL_SLEEP, CLASS_MAGIC_USER, 8);
	spell_level(SPELL_LIGHTNING_BOLT, CLASS_MAGIC_USER, 9);
	spell_level(SPELL_BLINDNESS, CLASS_MAGIC_USER, 9);
	spell_level(SPELL_DETECT_POISON, CLASS_MAGIC_USER, 10);
	spell_level(SPELL_FLY, CLASS_MAGIC_USER, 10);
	spell_level(SPELL_COLOR_SPRAY, CLASS_MAGIC_USER, 11);
	spell_level(SPELL_MANA_TRANSFER, CLASS_MAGIC_USER, 11);
	spell_level(SPELL_TELEPORT, CLASS_MAGIC_USER, 12);
	spell_level(SPELL_ENERGY_DRAIN, CLASS_MAGIC_USER, 13);
	spell_level(SPELL_LIFE_TRANSFER, CLASS_MAGIC_USER, 13);
	spell_level(SPELL_CURSE, CLASS_MAGIC_USER, 14);
	spell_level(SPELL_POISON, CLASS_MAGIC_USER, 14);
	spell_level(SPELL_FIREBALL, CLASS_MAGIC_USER, 15);
	spell_level(SPELL_SCRY, CLASS_MAGIC_USER, 15);
	spell_level(SPELL_CHARM, CLASS_MAGIC_USER, 16);
	spell_level(SPELL_SENSE_LIFE, CLASS_MAGIC_USER, 17);
	spell_level(SPELL_MIRROR_IMAGE, CLASS_MAGIC_USER, 18); 
	spell_level(SPELL_RELOCATE, CLASS_MAGIC_USER, 19);
	spell_level(SPELL_WALL_OF_FOG, CLASS_MAGIC_USER, 20);
	spell_level(SPELL_RECHARGE, CLASS_MAGIC_USER, 22);
	spell_level(SPELL_PORTAL, CLASS_MAGIC_USER, 24);
	spell_level(SPELL_ENCHANT_WEAPON, CLASS_MAGIC_USER, 26);
	spell_level(SPELL_DISINTEGRATE, CLASS_MAGIC_USER, 28);
	spell_level(SPELL_CLONE, CLASS_MAGIC_USER, 30);

	spell_level(SKILL_BANDAGE, CLASS_MAGIC_USER, 1);
	spell_level(SKILL_MOUNT, CLASS_MAGIC_USER, 1);
	spell_level(SKILL_RIDING, CLASS_MAGIC_USER, 1);
	spell_level(SKILL_BREW, CLASS_MAGIC_USER, 10);
	spell_level(SKILL_SCRIBE, CLASS_MAGIC_USER, 15);
	spell_level(SKILL_TAME, CLASS_MAGIC_USER, 25);

	/* CLERICS */
	spell_level(SPELL_CURE_LIGHT, CLASS_CLERIC, 1);
	spell_level(SPELL_ARMOR, CLASS_CLERIC, 1);
	spell_level(SPELL_CREATE_FOOD, CLASS_CLERIC, 2);
	spell_level(SPELL_CREATE_WATER, CLASS_CLERIC, 2);
	spell_level(SPELL_DETECT_POISON, CLASS_CLERIC, 3);
	spell_level(SPELL_AWARENESS, CLASS_CLERIC, 3);
	spell_level(SPELL_DETECT_ALIGN, CLASS_CLERIC, 4);
	spell_level(SPELL_CURE_BLIND, CLASS_CLERIC, 4);
	spell_level(SPELL_BLESS, CLASS_CLERIC, 5);
	spell_level(SPELL_DETECT_INVIS, CLASS_CLERIC, 6);
	spell_level(SPELL_BLINDNESS, CLASS_CLERIC, 6);
	spell_level(SPELL_INFRAVISION, CLASS_CLERIC, 7);
	spell_level(SPELL_PROT_FROM_EVIL, CLASS_CLERIC, 8);
	spell_level(SPELL_POISON, CLASS_CLERIC, 8);
	spell_level(SPELL_GROUP_ARMOR, CLASS_CLERIC, 9);
	spell_level(SPELL_CURE_CRITIC, CLASS_CLERIC, 9);
	spell_level(SPELL_SUMMON, CLASS_CLERIC, 10);
	spell_level(SPELL_REMOVE_POISON, CLASS_CLERIC, 10);
	spell_level(SPELL_LIFE_TRANSFER, CLASS_CLERIC, 11);
	spell_level(SPELL_WORD_OF_RECALL, CLASS_CLERIC, 12);
	spell_level(SPELL_EARTHQUAKE, CLASS_CLERIC, 12);
	spell_level(SPELL_MANA_TRANSFER, CLASS_CLERIC, 13);
	spell_level(SPELL_DISPEL_EVIL, CLASS_CLERIC, 14);
	spell_level(SPELL_DISPEL_GOOD, CLASS_CLERIC, 14);
	spell_level(SPELL_SANCTUARY, CLASS_CLERIC, 15);
	spell_level(SPELL_CALL_LIGHTNING, CLASS_CLERIC, 15);
	spell_level(SPELL_HEAL, CLASS_CLERIC, 16);
	spell_level(SPELL_CONTROL_WEATHER, CLASS_CLERIC, 17);
	spell_level(SPELL_SENSE_LIFE, CLASS_CLERIC, 18);
	spell_level(SPELL_HARM, CLASS_CLERIC, 19);
	spell_level(SPELL_RECHARGE, CLASS_CLERIC, 20);
	spell_level(SPELL_GROUP_HEAL, CLASS_CLERIC, 22);
	spell_level(SPELL_REGEN, CLASS_CLERIC, 24);
	spell_level(SPELL_REMOVE_CURSE, CLASS_CLERIC, 26);
	spell_level(SPELL_GROUP_RECALL, CLASS_CLERIC, 30);

	spell_level(SKILL_BANDAGE, CLASS_CLERIC, 1);
	spell_level(SKILL_MOUNT, CLASS_CLERIC, 1);
	spell_level(SKILL_RIDING, CLASS_CLERIC, 1);
	spell_level(SKILL_BREW, CLASS_CLERIC, 10);
	spell_level(SKILL_SCRIBE, CLASS_CLERIC, 15);
	spell_level(SKILL_TAME, CLASS_CLERIC, 18);
   spell_level(SKILL_SECOND_ATTACK, CLASS_CLERIC, 28);

	/* THIEVES */
	spell_level(SKILL_BANDAGE, CLASS_THIEF, 1);
	spell_level(SKILL_HIDE, CLASS_THIEF, 1);
	spell_level(SKILL_MOUNT, CLASS_THIEF, 1);
	spell_level(SKILL_RIDING, CLASS_THIEF, 1);
	spell_level(SKILL_PICK_LOCK, CLASS_THIEF, 2);
	spell_level(SKILL_SAP, CLASS_THIEF, 2);
	spell_level(SKILL_BACKSTAB, CLASS_THIEF, 3);
	spell_level(SKILL_LUCK, CLASS_THIEF, 3);
	spell_level(SKILL_STEAL, CLASS_THIEF, 4);
	spell_level(SKILL_SNEAK, CLASS_THIEF, 5);
	spell_level(SKILL_THROW, CLASS_THIEF, 6);
	spell_level(SKILL_TRACK, CLASS_THIEF, 6);
	spell_level(SKILL_LISTEN, CLASS_THIEF, 7);
	spell_level(SKILL_FARSCAN, CLASS_THIEF, 8);
	spell_level(SKILL_SCOUT, CLASS_THIEF, 9);
	spell_level(SKILL_TAME, CLASS_THIEF, 10);
   spell_level(SKILL_SECOND_ATTACK, CLASS_THIEF, 12);
	spell_level(SKILL_POISON, CLASS_THIEF, 15);
	spell_level(SKILL_DUAL_WIELD, CLASS_THIEF, 20);
   spell_level(SKILL_THIRD_ATTACK, CLASS_THIEF, 24);

	/* WARRIORS */
	spell_level(SKILL_BANDAGE, CLASS_WARRIOR, 1);
	spell_level(SKILL_KICK, CLASS_WARRIOR, 1);
	spell_level(SKILL_MOUNT, CLASS_WARRIOR, 1);
	spell_level(SKILL_RIDING, CLASS_WARRIOR, 1);
	spell_level(SKILL_RETREAT, CLASS_WARRIOR, 2);
	spell_level(SKILL_RESCUE, CLASS_WARRIOR, 3);
	spell_level(SKILL_THROW, CLASS_WARRIOR, 3);
	spell_level(SKILL_TAME, CLASS_WARRIOR, 4);
	spell_level(SKILL_DUAL_WIELD, CLASS_WARRIOR, 5);
	spell_level(SKILL_FORAGE, CLASS_WARRIOR, 6);
	spell_level(SKILL_UNARMED_COMBAT, CLASS_WARRIOR, 6);
	spell_level(SKILL_DISARM, CLASS_WARRIOR, 7);
   spell_level(SKILL_SECOND_ATTACK, CLASS_WARRIOR, 8);
	spell_level(SKILL_TRACK, CLASS_WARRIOR, 9);
	spell_level(SKILL_TAN, CLASS_WARRIOR, 10);
	spell_level(SKILL_BASH, CLASS_WARRIOR, 12);
	spell_level(SKILL_STUN, CLASS_WARRIOR, 15);
   spell_level(SKILL_THIRD_ATTACK, CLASS_WARRIOR, 16);
	spell_level(SKILL_ENHANCED_DMG, CLASS_WARRIOR, 18);
	spell_level(SKILL_FORGE, CLASS_WARRIOR, 20);
	spell_level(SKILL_GUT, CLASS_WARRIOR, 22);
	spell_level(SKILL_WHIRLWIND, CLASS_WARRIOR, 25);
}


/*
 * This is the exp given to implementors -- it must always be greater
 * than the exp required for immortality, plus at least 20,000 or so.
 */
#define EXP_MAX  10000000

/* Function to return the exp required for each class/level */
int level_exp(int chclass, int level)
{
	if (level > LVL_IMPL || level < 0) {
		log("SYSERR: Requesting exp for invalid level %d!", level);
		return 0;
	}

	/*
	 * Gods have exp close to EXP_MAX.  This statement should never have to
	 * changed, regardless of how many mortal or immortal levels exist.
	 */
   if (level > LVL_IMMORT) {
		return EXP_MAX - ((LVL_IMPL - level) * 1000);
   }

	/* Exp required for normal mortals is below */

	switch (chclass) {

		case CLASS_MAGIC_USER:
			switch (level) {
				case  0: return 0;
				case  1: return 1;
				case  2: return 2500;
				case  3: return 5000;
				case  4: return 10000;
				case  5: return 20000;
				case  6: return 40000;
				case  7: return 60000;
				case  8: return 90000;
				case  9: return 135000;
				case 10: return 250000;
				case 11: return 375000;
				case 12: return 750000;
				case 13: return 1125000;
				case 14: return 1500000;
				case 15: return 1875000;
				case 16: return 2250000;
				case 17: return 2625000;
				case 18: return 3000000;
				case 19: return 3375000;
				case 20: return 3750000;
				case 21: return 4000000;
				case 22: return 4300000;
				case 23: return 4600000;
				case 24: return 4900000;
				case 25: return 5200000;
				case 26: return 5500000;
				case 27: return 5950000;
				case 28: return 6400000;
				case 29: return 6850000;
				case 30: return 7400000;
					/* add new levels here */
				case LVL_IMMORT: return 8000000;
			}
			break;

		case CLASS_CLERIC:
			switch (level) {
				case  0: return 0;
				case  1: return 1;
				case  2: return 1500;
				case  3: return 3000;
				case  4: return 6000;
				case  5: return 13000;
				case  6: return 27500;
				case  7: return 55000;
				case  8: return 110000;
				case  9: return 225000;
				case 10: return 450000;
				case 11: return 675000;
				case 12: return 900000;
				case 13: return 1125000;
				case 14: return 1350000;
				case 15: return 1575000;
				case 16: return 1800000;
				case 17: return 2100000;
				case 18: return 2400000;
				case 19: return 2700000;
				case 20: return 3000000;
				case 21: return 3250000;
				case 22: return 3500000;
				case 23: return 3800000;
				case 24: return 4100000;
				case 25: return 4400000;
				case 26: return 4800000;
				case 27: return 5200000;
				case 28: return 5600000;
				case 29: return 6000000;
				case 30: return 6400000;
					/* add new levels here */
				case LVL_IMMORT: return 7000000;
			}
			break;

		case CLASS_THIEF:
			switch (level) {
				case  0: return 0;
				case  1: return 1;
				case  2: return 1250;
				case  3: return 2500;
				case  4: return 5000;
				case  5: return 10000;
				case  6: return 20000;
				case  7: return 30000;
				case  8: return 70000;
				case  9: return 110000;
				case 10: return 160000;
				case 11: return 220000;
				case 12: return 440000;
				case 13: return 660000;
				case 14: return 880000;
				case 15: return 1100000;
				case 16: return 1500000;
				case 17: return 2000000;
				case 18: return 2500000;
				case 19: return 3000000;
				case 20: return 3500000;
				case 21: return 3650000;
				case 22: return 3800000;
				case 23: return 4100000;
				case 24: return 4400000;
				case 25: return 4700000;
				case 26: return 5100000;
				case 27: return 5500000;
				case 28: return 5900000;
				case 29: return 6300000;
				case 30: return 6650000;
					/* add new levels here */
				case LVL_IMMORT: return 7000000;
			}
			break;

		case CLASS_WARRIOR:
			switch (level) {
				case  0: return 0;
				case  1: return 1;
				case  2: return 2000;
				case  3: return 4000;
				case  4: return 8000;
				case  5: return 16000;
				case  6: return 32000;
				case  7: return 64000;
				case  8: return 125000;
				case  9: return 250000;
				case 10: return 500000;
				case 11: return 750000;
				case 12: return 1000000;
				case 13: return 1250000;
				case 14: return 1500000;
				case 15: return 1850000;
				case 16: return 2200000;
				case 17: return 2550000;
				case 18: return 2900000;
				case 19: return 3250000;
				case 20: return 3600000;
				case 21: return 3900000;
				case 22: return 4200000;
				case 23: return 4500000;
				case 24: return 4800000;
				case 25: return 5150000;
				case 26: return 5500000;
				case 27: return 5950000;
				case 28: return 6400000;
				case 29: return 6850000;
				case 30: return 7400000;
					/* add new levels here */
				case LVL_IMMORT: return 8000000;
			}
			break;
	}

	/*
	 * This statement should never be reached if the exp tables in this function
	 * are set up properly.  If you see exp of 123456 then the tables above are
	 * incomplete -- so, complete them!
	 */
	log("SYSERR: XP tables not set up correctly in class.c!");
	return 123456;
}


/* 
 * Default titles of male characters.
 */
const char *title_male(int chclass, int level)
{
	if (level <= 0 || level > LVL_IMPL)
		return "the Man";
	if (level == LVL_IMPL)
		return "the Implementor";

	switch (chclass) {

		case CLASS_MAGIC_USER:
			switch (level) {
				case  1: return "the Apprentice of Magic";
				case  2: return "the Spell Student";
				case  3: return "the Scholar of Magic";
				case  4: return "the Delver in Spells";
				case  5: return "the Medium of Magic";
				case  6: return "the Scribe of Magic";
				case  7: return "the Seer";
				case  8: return "the Sage";
				case  9: return "the Illusionist";
				case 10: return "the Abjurer";
				case 11: return "the Invoker";
				case 12: return "the Enchanter";
				case 13: return "the Conjurer";
				case 14: return "the Magician";
				case 15: return "the Creator";
				case 16: return "the Savant";
				case 17: return "the Magus";
				case 18: return "the Wizard";
				case 19: return "the Warlock";
				case 20: return "the Sorcerer";
				case 21: return "the Necromancer";
				case 22: return "the Thaumaturge";
				case 23: return "the Student of the Occult";
				case 24: return "the Disciple of the Uncanny";
				case 25: return "the Minor Elemental";
				case 26: return "the Greater Elemental";
				case 27: return "the Crafter of Magics";
				case 28: return "the Shaman";
				case 29: return "the Keeper of Talismans";
				case 30: return "the Archmage";
				case LVL_IMMORT: return "the Immortal Warlock";
				case LVL_GOD: return "the Avatar of Magic";
				case LVL_GRGOD: return "the God of Magic";
				default: return "the Mage";
			}
			break;

		case CLASS_CLERIC:
			switch (level) {
				case  1: return "the Believer";
				case  2: return "the Attendant";
				case  3: return "the Acolyte";
				case  4: return "the Novice";
				case  5: return "the Missionary";
				case  6: return "the Adept";
				case  7: return "the Deacon";
				case  8: return "the Vicar";
				case  9: return "the Priest";
				case 10: return "the Minister";
				case 11: return "the Canon";
				case 12: return "the Levite";
				case 13: return "the Curate";
				case 14: return "the Monk";
				case 15: return "the Healer";
				case 16: return "the Chaplain";
				case 17: return "the Expositor";
				case 18: return "the Bishop";
				case 19: return "the Arch Bishop";
				case 20: return "the Patriarch";
					/* no one ever thought up these titles 21-30 */
				case LVL_IMMORT: return "the Immortal Cardinal";
				case LVL_GOD: return "the Inquisitor";
				case LVL_GRGOD: return "the God of good and evil";
				default: return "the Cleric";
			}
			break;

		case CLASS_THIEF:
			switch (level) {
				case  1: return "the Pilferer";
				case  2: return "the Footpad";
				case  3: return "the Filcher";
				case  4: return "the Pick-Pocket";
				case  5: return "the Sneak";
				case  6: return "the Pincher";
				case  7: return "the Cut-Purse";
				case  8: return "the Snatcher";
				case  9: return "the Sharper";
				case 10: return "the Rogue";
				case 11: return "the Robber";
				case 12: return "the Magsman";
				case 13: return "the Highwayman";
				case 14: return "the Burglar";
				case 15: return "the Thief";
				case 16: return "the Knifer";
				case 17: return "the Quick-Blade";
				case 18: return "the Killer";
				case 19: return "the Brigand";
				case 20: return "the Cut-Throat";
					/* no one ever thought up these titles 21-30 */
				case LVL_IMMORT: return "the Immortal Assasin";
				case LVL_GOD: return "the Demi God of thieves";
				case LVL_GRGOD: return "the God of thieves and tradesmen";
				default: return "the Thief";
			}
			break;

		case CLASS_WARRIOR:
			switch(level) {
				case  1: return "the Swordpupil";
				case  2: return "the Recruit";
				case  3: return "the Sentry";
				case  4: return "the Fighter";
				case  5: return "the Soldier";
				case  6: return "the Warrior";
				case  7: return "the Veteran";
				case  8: return "the Swordsman";
				case  9: return "the Fencer";
				case 10: return "the Combatant";
				case 11: return "the Hero";
				case 12: return "the Myrmidon";
				case 13: return "the Swashbuckler";
				case 14: return "the Mercenary";
				case 15: return "the Swordmaster";
				case 16: return "the Lieutenant";
				case 17: return "the Champion";
				case 18: return "the Dragoon";
				case 19: return "the Cavalier";
				case 20: return "the Knight";
					/* no one ever thought up these titles 21-30 */
				case LVL_IMMORT: return "the Immortal Warlord";
				case LVL_GOD: return "the Extirpator";
				case LVL_GRGOD: return "the God of war";
				default: return "the Warrior";
			}
			break;
	}

	/* Default title for classes which do not have titles defined */
	return "the Classless";
}


/* 
 * Default titles of female characters.
 */
const char *title_female(int chclass, int level)
{
	if (level <= 0 || level > LVL_IMPL)
		return "the Woman";
	if (level == LVL_IMPL)
		return "the Implementress";

	switch (chclass) {

		case CLASS_MAGIC_USER:
			switch (level) {
				case  1: return "the Apprentice of Magic";
				case  2: return "the Spell Student";
				case  3: return "the Scholar of Magic";
				case  4: return "the Delveress in Spells";
				case  5: return "the Medium of Magic";
				case  6: return "the Scribess of Magic";
				case  7: return "the Seeress";
				case  8: return "the Sage";
				case  9: return "the Illusionist";
				case 10: return "the Abjuress";
				case 11: return "the Invoker";
				case 12: return "the Enchantress";
				case 13: return "the Conjuress";
				case 14: return "the Witch";
				case 15: return "the Creator";
				case 16: return "the Savant";
				case 17: return "the Craftess";
				case 18: return "the Wizard";
				case 19: return "the War Witch";
				case 20: return "the Sorceress";
				case 21: return "the Necromancress";
				case 22: return "the Thaumaturgess";
				case 23: return "the Student of the Occult";
				case 24: return "the Disciple of the Uncanny";
				case 25: return "the Minor Elementress";
				case 26: return "the Greater Elementress";
				case 27: return "the Crafter of Magics";
				case 28: return "Shaman";
				case 29: return "the Keeper of Talismans";
				case 30: return "Archwitch";
				case LVL_IMMORT: return "the Immortal Enchantress";
				case LVL_GOD: return "the Empress of Magic";
				case LVL_GRGOD: return "the Goddess of Magic";
				default: return "the Witch";
			}
			break;

		case CLASS_CLERIC:
			switch (level) {
				case  1: return "the Believer";
				case  2: return "the Attendant";
				case  3: return "the Acolyte";
				case  4: return "the Novice";
				case  5: return "the Missionary";
				case  6: return "the Adept";
				case  7: return "the Deaconess";
				case  8: return "the Vicaress";
				case  9: return "the Priestess";
				case 10: return "the Lady Minister";
				case 11: return "the Canon";
				case 12: return "the Levitess";
				case 13: return "the Curess";
				case 14: return "the Nunne";
				case 15: return "the Healess";
				case 16: return "the Chaplain";
				case 17: return "the Expositress";
				case 18: return "the Bishop";
				case 19: return "the Arch Lady of the Church";
				case 20: return "the Matriarch";
					/* no one ever thought up these titles 21-30 */
				case LVL_IMMORT: return "the Immortal Priestess";
				case LVL_GOD: return "the Inquisitress";
				case LVL_GRGOD: return "the Goddess of good and evil";
				default: return "the Cleric";
			}
			break;

		case CLASS_THIEF:
			switch (level) {
				case  1: return "the Pilferess";
				case  2: return "the Footpad";
				case  3: return "the Filcheress";
				case  4: return "the Pick-Pocket";
				case  5: return "the Sneak";
				case  6: return "the Pincheress";
				case  7: return "the Cut-Purse";
				case  8: return "the Snatcheress";
				case  9: return "the Sharpress";
				case 10: return "the Rogue";
				case 11: return "the Robber";
				case 12: return "the Magswoman";
				case 13: return "the Highwaywoman";
				case 14: return "the Burglaress";
				case 15: return "the Thief";
				case 16: return "the Knifer";
				case 17: return "the Quick-Blade";
				case 18: return "the Murderess";
				case 19: return "the Brigand";
				case 20: return "the Cut-Throat";
					/* no one ever thought up these titles 21-30 */
				case LVL_IMMORT: return "the Immortal Assasin";
				case LVL_GOD: return "the Demi Goddess of thieves";
				case LVL_GRGOD: return "the Goddess of thieves and tradesmen";
				default: return "the Thief";
			}
			break;

		case CLASS_WARRIOR:
			switch(level) {
				case  1: return "the Swordpupil";
				case  2: return "the Recruit";
				case  3: return "the Sentress";
				case  4: return "the Fighter";
				case  5: return "the Soldier";
				case  6: return "the Warrior";
				case  7: return "the Veteran";
				case  8: return "the Swordswoman";
				case  9: return "the Fenceress";
				case 10: return "the Combatess";
				case 11: return "the Heroine";
				case 12: return "the Myrmidon";
				case 13: return "the Swashbuckleress";
				case 14: return "the Mercenaress";
				case 15: return "the Swordmistress";
				case 16: return "the Lieutenant";
				case 17: return "the Lady Champion";
				case 18: return "the Lady Dragoon";
				case 19: return "the Cavalier";
				case 20: return "the Lady Knight";
					/* no one ever thought up these titles 21-30 */
				case LVL_IMMORT: return "the Immortal Lady of War";
				case LVL_GOD: return "the Queen of Destruction";
				case LVL_GRGOD: return "the Goddess of war";
				default: return "the Warrior";
			}
			break;
	}

	/* Default title for classes which do not have titles defined */
	return "the Classless";
}

