/* ************************************************************************
 *   File: db.c                                          Part of CircleMUD *
 *  Usage: Loading/saving chars, booting/resetting world, internal funcs   *
 *                                                                         *
 *  All rights reserved.  See license.doc for complete information.        *
 *                                                                         *
 *  Copyright (C) 1993, 94 by the Trustees of the Johns Hopkins University *
 *  CircleMUD is based on DikuMUD, Copyright (C) 1990, 1991.               *
 ************************************************************************ */

#define __DB_C__

#include "conf.h"
#include "sysdep.h"

#include "structs.h"
#include "utils.h"

#include "comm.h"
#include "constants.h"
#include "db.h"
#include "handler.h"
#include "house.h"
#include "interpreter.h"
#include "mail.h"
#include "spells.h"


/* Extern and forward declarations */
void clean_pfiles(void);

extern int auto_pwipe;


/**************************************************************************
 *  declarations of most of the 'global' variables                         *
 **************************************************************************/
void init_mm (void);

struct raff_node *raff_list = NULL;  /* list of room affections */

struct room_data *world = NULL;	/* array of rooms		 */
room_rnum top_of_world = 0;	/* ref to top element of world	 */

struct char_data *character_list = NULL;	/* global linked list of
														 * chars	 */
struct index_data *mob_index;	/* index table for mobile file	 */
struct char_data *mob_proto;	/* prototypes for mobs		 */
mob_rnum top_of_mobt = 0;	/* top of mobile index table	 */

struct obj_data *object_list = NULL;	/* global linked list of objs	 */
struct index_data *obj_index;	/* index table for object file	 */
struct obj_data *obj_proto;	/* prototypes for objs		 */
obj_rnum top_of_objt = 0;	/* top of object index table	 */

struct zone_data *zone_table;	/* zone table			 */
zone_rnum top_of_zone_table = 0;/* top element of zone tab	 */
struct message_list fight_messages[MAX_MESSAGES];	/* fighting messages	 */

int no_mail = 0;		/* mail disabled?		 */
int mini_mud = 0;		/* mini-mud mode?		 */
int no_rent_check = 0;		/* skip rent check on boot?	 */
time_t boot_time = 0;		/* time of mud boot		 */
int circle_restrict = 0;	/* level of game restriction	 */

char *credits = NULL;		/* game credits			 */
char *news = NULL;		/* mud news			 */
char *motd = NULL;		/* message of the day - mortals */
char *imotd = NULL;		/* message of the day - immorts */
char *GREETINGS = NULL;		/* opening credits screen	*/
char *help = NULL;		/* help screen			 */
char *info = NULL;		/* info page			 */
char *wizlist = NULL;		/* list of higher gods		 */
char *immlist = NULL;		/* list of peon gods		 */
char *background = NULL;	/* background story		 */
char *handbook = NULL;		/* handbook for new immortals	 */
char *policies = NULL;		/* policies page		 */
obj_vnum *food_list = NULL;
int num_foods = 0;

struct help_index_element *help_table = 0;	/* the help table	 */
int top_of_helpt = 0;		/* top of help index table	 */

struct time_info_data time_info;/* the infomation about the time    */
struct player_special_data dummy_mob;	/* dummy spec area for mobs	*/
struct reset_q_type reset_q;	/* queue of zones to be reset	 */


/* extern functions */
void parse_room(FILE *fl, int virtual_nr);
char *parse_object(FILE *obj_f, int nr);
void assign_the_gms(void);
void boot_the_guilds(FILE *gm_f, char *filename, int rec_count);


/* local functions */
void index_boot(int mode);
void discrete_load(FILE *fl, int mode, char *filename);
void parse_mobile(FILE *mob_f, int nr);
void load_zones(FILE *fl, char *zonename);
void load_help(FILE *fl);
void assign_mobiles(void);
void assign_objects(void);
void assign_rooms(void);
void assign_the_shopkeepers(void);
void build_player_index(void);
void reset_zone(zone_rnum zone);
int file_to_string(const char *name, char *buf);
int file_to_string_alloc(const char *name, char **buf);
ACMD(do_reboot);
void boot_world(void);
int count_alias_records(FILE *fl);
int count_hash_records(FILE *fl);
void get_one_line(FILE *fl, char *buf);
void check_start_rooms(void);
void renum_world(void);
void renum_zone_table(void);
void reset_time(void);
void load_bank(void);

/* external functions */
void paginate_string(char *str, struct descriptor_data *d);
struct time_info_data *mud_time_passed(time_t t2, time_t t1);
void load_messages(void);
void weather_and_time(int mode);
void mag_assign_spells(void);
void boot_social_messages(void);
void update_obj_file(void);	/* In objsave.c */
void sort_commands(void);
void sort_spells(void);
void load_banned(void);
void Read_Invalid_List(void);
void boot_the_shops(FILE *shop_f, char *filename, int rec_count);
int hsort(const void *a, const void *b);
void prune_crlf(char *txt);
void destroy_shops(void);

/* external vars */
extern int no_specials;
extern struct descriptor_data *descriptor_list;

/*************************************************************************
 *  routines for booting the system                                       *
 *************************************************************************/

/* Wipe out all the loaded text files, for shutting down. */
void free_text_files(void)
{
	char **textfiles[] = {
		&wizlist, &immlist, &news, &credits, &motd, &imotd, &help, &info,
		&policies, &handbook, &background, &GREETINGS, NULL
	};
	int rf;

	for (rf = 0; textfiles[rf]; rf++)
		if (*textfiles[rf]) {
			free(*textfiles[rf]);
			*textfiles[rf] = NULL;
		}
}


void init_food_list(void)
{
	obj_rnum i;
	int cnt;

	/* First we have to count how many food items so we can allocate
	 * only as much space as need for the food array. */
	for (i = 0; i < top_of_objt; i++) {
		if (GET_OBJ_TYPE(&obj_proto[i]) == ITEM_FOOD)
			num_foods += 1;
	}

	/* Now allocate */
	if (num_foods) {
		CREATE(food_list, obj_vnum, num_foods);
		if (!food_list)
			return;
	
		/* And now go through and assign the values to the array */
		for (i = 0, cnt = 0; i < top_of_objt; i++) {
			if (GET_OBJ_TYPE(&obj_proto[i]) == ITEM_FOOD)
				food_list[cnt++] = GET_OBJ_VNUM(&obj_proto[i]);
		}
	}
}


/*
 * Too bad it doesn't check the return values to let the user
 * know about -1 values.  This will result in an 'Okay.' to a
 * 'reload' command even when the string was not replaced.
 * To fix later, if desired. -gg 6/24/99
 */
ACMD(do_reboot)
{
	char arg[MAX_INPUT_LENGTH];

	one_argument(argument, arg);

	if (!str_cmp(arg, "all") || *arg == '*') {
		if (file_to_string_alloc(GREETINGS_FILE, &GREETINGS) == 0)
			prune_crlf(GREETINGS);
		file_to_string_alloc(WIZLIST_FILE, &wizlist);
		file_to_string_alloc(IMMLIST_FILE, &immlist);
		file_to_string_alloc(NEWS_FILE, &news);
		file_to_string_alloc(CREDITS_FILE, &credits);
		file_to_string_alloc(MOTD_FILE, &motd);
		file_to_string_alloc(IMOTD_FILE, &imotd);
		file_to_string_alloc(HELP_PAGE_FILE, &help);
		file_to_string_alloc(INFO_FILE, &info);
		file_to_string_alloc(POLICIES_FILE, &policies);
		file_to_string_alloc(HANDBOOK_FILE, &handbook);
		file_to_string_alloc(BACKGROUND_FILE, &background);
	} else if (!str_cmp(arg, "wizlist"))
		file_to_string_alloc(WIZLIST_FILE, &wizlist);
	else if (!str_cmp(arg, "immlist"))
		file_to_string_alloc(IMMLIST_FILE, &immlist);
	else if (!str_cmp(arg, "news"))
		file_to_string_alloc(NEWS_FILE, &news);
	else if (!str_cmp(arg, "credits"))
		file_to_string_alloc(CREDITS_FILE, &credits);
	else if (!str_cmp(arg, "motd"))
		file_to_string_alloc(MOTD_FILE, &motd);
	else if (!str_cmp(arg, "imotd"))
		file_to_string_alloc(IMOTD_FILE, &imotd);
	else if (!str_cmp(arg, "help"))
		file_to_string_alloc(HELP_PAGE_FILE, &help);
	else if (!str_cmp(arg, "info"))
		file_to_string_alloc(INFO_FILE, &info);
	else if (!str_cmp(arg, "policy"))
		file_to_string_alloc(POLICIES_FILE, &policies);
	else if (!str_cmp(arg, "handbook"))
		file_to_string_alloc(HANDBOOK_FILE, &handbook);
	else if (!str_cmp(arg, "background"))
		file_to_string_alloc(BACKGROUND_FILE, &background);
	else if (!str_cmp(arg, "greetings")) {
		if (file_to_string_alloc(GREETINGS_FILE, &GREETINGS) == 0)
			prune_crlf(GREETINGS);
	} else if (!str_cmp(arg, "xhelp")) {
		if (help_table)
			free_help();
		index_boot(DB_BOOT_HLP);
	} else {
		send_to_char(ch, "Unknown reload option.\r\n");
		return;
	}

	send_to_char(ch, "%s", OK);
}


void boot_world(void)
{
	log("Loading zone table.");
	index_boot(DB_BOOT_ZON);

	log("Loading rooms.");
	index_boot(DB_BOOT_WLD);

	log("Renumbering rooms.");
	renum_world();

	log("Checking start rooms.");
	check_start_rooms();

	log("Loading mobs and generating index.");
	index_boot(DB_BOOT_MOB);

	log("Loading objs and generating index.");
	index_boot(DB_BOOT_OBJ);

	log("Renumbering zone table.");
	renum_zone_table();

	if (!no_specials) {
		log("Loading shops.");
		index_boot(DB_BOOT_SHP);

		log("Loading guild masters.");
		index_boot(DB_BOOT_GLD);
	}

	log("Initializing statics.");
	init_food_list();
}


void free_extra_descriptions(struct extra_descr_data *edesc)
{
	struct extra_descr_data *enext;

	for (; edesc; edesc = enext) {
		enext = edesc->next;

		free(edesc->keyword);
		free(edesc->description);
		free(edesc);
	}
}


/* Free the world, in a memory allocation sense. */
void destroy_db(void)
{
	ssize_t cnt, itr;
	struct char_data *chtmp;
	struct obj_data *objtmp;

	/* Active Mobiles & Players */
	while (character_list) {
		chtmp = character_list;
		character_list = character_list->next;
		free_char(chtmp);
	}

	/* Active Objects */
	while (object_list) {
		objtmp = object_list;
		object_list = object_list->next;
		free_obj(objtmp);
	}

	/* Rooms */
	for (cnt = 0; cnt <= top_of_world; cnt++) {
		if (world[cnt].name)
			free(world[cnt].name);
		if (world[cnt].description)
			free(world[cnt].description);
		free_extra_descriptions(world[cnt].ex_description);

		for (itr = 0; itr < NUM_OF_DIRS; itr++) {
			if (!world[cnt].dir_option[itr])
				continue;

			if (world[cnt].dir_option[itr]->general_description)
				free(world[cnt].dir_option[itr]->general_description);
			if (world[cnt].dir_option[itr]->keyword)
				free(world[cnt].dir_option[itr]->keyword);
			free(world[cnt].dir_option[itr]);
		}
	}
	free(world);

	/* Objects */
	for (cnt = 0; cnt <= top_of_objt; cnt++) {
		if (obj_proto[cnt].name)
			free(obj_proto[cnt].name);
		if (obj_proto[cnt].description)
			free(obj_proto[cnt].description);
		if (obj_proto[cnt].short_description)
			free(obj_proto[cnt].short_description);
		if (obj_proto[cnt].action_description)
			free(obj_proto[cnt].action_description);
		free_extra_descriptions(obj_proto[cnt].ex_description);
	}
	free(obj_proto);
	free(obj_index);

	if (food_list)
		free(food_list);

	/* Mobiles */
	for (cnt = 0; cnt <= top_of_mobt; cnt++) {
		if (mob_proto[cnt].player.name)
			free(mob_proto[cnt].player.name);
		if (mob_proto[cnt].player.title)
			free(mob_proto[cnt].player.title);
		if (mob_proto[cnt].player.short_descr)
			free(mob_proto[cnt].player.short_descr);
		if (mob_proto[cnt].player.long_descr)
			free(mob_proto[cnt].player.long_descr);
		if (mob_proto[cnt].player.description)
			free(mob_proto[cnt].player.description);

		while (mob_proto[cnt].affected)
			affect_remove(&mob_proto[cnt], mob_proto[cnt].affected);
	}
	free(mob_proto);
	free(mob_index);

	/* Shops */
	destroy_shops();

	/* Zones */
	for (cnt = 0; cnt <= top_of_zone_table; cnt++) {
		if (zone_table[cnt].name)
			free(zone_table[cnt].name);
		if (zone_table[cnt].cmd)
			free(zone_table[cnt].cmd);
	}
	free(zone_table);
}


/* body of the booting system */
void boot_db(void)
{
	zone_rnum i;

	log("Boot db -- BEGIN.");

	log("Resetting the game time:");
	reset_time();

	/* init random number generator */
	init_mm();

	log("Reading news, credits, help, bground, info & motds.");
	file_to_string_alloc(NEWS_FILE, &news);
	file_to_string_alloc(CREDITS_FILE, &credits);
	file_to_string_alloc(MOTD_FILE, &motd);
	file_to_string_alloc(IMOTD_FILE, &imotd);
	file_to_string_alloc(HELP_PAGE_FILE, &help);
	file_to_string_alloc(INFO_FILE, &info);
	file_to_string_alloc(WIZLIST_FILE, &wizlist);
	file_to_string_alloc(IMMLIST_FILE, &immlist);
	file_to_string_alloc(POLICIES_FILE, &policies);
	file_to_string_alloc(HANDBOOK_FILE, &handbook);
	file_to_string_alloc(BACKGROUND_FILE, &background);
	if (file_to_string_alloc(GREETINGS_FILE, &GREETINGS) == 0)
		prune_crlf(GREETINGS);

	log("Loading spell definitions.");
	mag_assign_spells();

	boot_world();

	log("Loading help entries.");
	index_boot(DB_BOOT_HLP);

	log("Generating player index.");
	build_player_index();

	if (auto_pwipe) {
		log("Cleaning out the pfiles.");
		clean_pfiles();
	}

	log("Loading fight messages.");
	load_messages();

	log("Loading social messages.");
	boot_social_messages();

	log("Assigning function pointers:");

	if (!no_specials) {
		log("   Mobiles.");
		assign_mobiles();
		log("   Shopkeepers.");
		assign_the_shopkeepers();
		log("   Guildmasters.");
		assign_the_gms();
		log("   Objects.");
		assign_objects();
		log("   Rooms.");
		assign_rooms();
	}

	log("Assigning spell and skill levels.");
	init_spell_levels();

	log("Sorting command list and spells.");
	sort_commands();
	sort_spells();

	log("Booting mail system.");
	if (!scan_file()) {
		log("    Mail boot failed -- Mail system disabled");
		no_mail = 1;
	}
	log("Reading banned site and invalid-name list.");
	load_banned();
	Read_Invalid_List();

	log("Loading bank.");
	load_bank();

	if (!no_rent_check) {
		log("Deleting timed-out crash and rent files:");
		update_obj_file();
		log("   Done.");
	}

	/* Moved here so the object limit code works. -gg 6/24/98 */
	if (!mini_mud) {
		log("Booting houses.");
		House_boot();
	}

	for (i = 0; i <= top_of_zone_table; i++) {
		log("Resetting #%d: %s (rooms %d-%d).", zone_table[i].number,
			 zone_table[i].name, zone_table[i].bot, zone_table[i].top);
		reset_zone(i);
	}

	reset_q.head = reset_q.tail = NULL;

	boot_time = time(0);
	MOBTrigger = TRUE;

	log("Boot db -- DONE.");
}


/* reset the time in the game from file */
void reset_time(void)
{
	time_t beginning_of_time = 0;
	FILE *bgtime;

	if ((bgtime = fopen(TIME_FILE, "r")) == NULL)
		log("SYSERR: Can't read from '%s' time file.", TIME_FILE);
	else {
		fscanf(bgtime, "%ld\n", &beginning_of_time);
		fclose(bgtime);
	}
	if (beginning_of_time == 0)
		beginning_of_time = 650336715;

	time_info = *mud_time_passed(time(0), beginning_of_time);

	if (time_info.hours <= 4)
		sunlight = SUN_DARK;
	else if (time_info.hours == 5)
		sunlight = SUN_RISE;
	else if (time_info.hours <= 20)
		sunlight = SUN_LIGHT;
	else if (time_info.hours == 21)
		sunlight = SUN_SET;
	else
		sunlight = SUN_DARK;

	log("   Current Gametime: %dH %dD %dM %dY.", time_info.hours,
		 time_info.day, time_info.month, time_info.year);
}


/* Write the time in 'when' to the MUD-time file. */
void save_mud_time(struct time_info_data *when)
{
	FILE *bgtime;

	if ((bgtime = fopen(TIME_FILE, "w")) == NULL)
		log("SYSERR: Can't write to '%s' time file.", TIME_FILE);
	else {
		fprintf(bgtime, "%ld\n", mud_time_to_secs(when));
		fclose(bgtime);
	}
}


/* function to count how many hash-mark delimited records exist in a file */
int count_hash_records(FILE *fl)
{
	char buf[128];
	int count = 0;

	while (fgets(buf, 128, fl))
		if (*buf == '#')
			count++;

	return (count);
}


void index_boot(int mode)
{
	const char *index_filename, *prefix = NULL;	/* NULL or egcs 1.1 complains */
	FILE *db_index, *db_file;
	int rec_count = 0, size[2];
	char buf2[PATH_MAX], buf1[MAX_STRING_LENGTH];

	switch (mode) {
		case DB_BOOT_WLD:
			prefix = WLD_PREFIX;
			break;
		case DB_BOOT_MOB:
			prefix = MOB_PREFIX;
			break;
		case DB_BOOT_OBJ:
			prefix = OBJ_PREFIX;
			break;
		case DB_BOOT_ZON:
			prefix = ZON_PREFIX;
			break;
		case DB_BOOT_SHP:
			prefix = SHP_PREFIX;
			break;
		case DB_BOOT_GLD:
			prefix = GLD_PREFIX;
			break;
		case DB_BOOT_PRG:
			prefix = PRG_PREFIX;
			break;
		case DB_BOOT_HLP:
			prefix = HLP_PREFIX;
			break;
		default:
			log("SYSERR: Unknown subcommand %d to index_boot!", mode);
			exit(1);
	}

	if (mini_mud)
		index_filename = MINDEX_FILE;
	else
		index_filename = INDEX_FILE;

	snprintf(buf2, sizeof(buf2), "%s%s", prefix, index_filename);
	if (!(db_index = fopen(buf2, "r"))) {
		log("SYSERR: opening index file '%s': %s", buf2, strerror(errno));
		exit(1);
	}

	/* first, count the number of records in the file so we can malloc */
	fscanf(db_index, "%s\n", buf1);
	while (*buf1 != '$') {
		snprintf(buf2, sizeof(buf2), "%s%s", prefix, buf1);
		if (!(db_file = fopen(buf2, "r"))) {
			log("SYSERR: File '%s' listed in '%s/%s': %s", buf2, prefix,
				 index_filename, strerror(errno));
			fscanf(db_index, "%s\n", buf1);
			continue;
		} else {
			if (mode == DB_BOOT_ZON)
				rec_count++;
			else if (mode == DB_BOOT_HLP)
				rec_count += count_alias_records(db_file);
			else
				rec_count += count_hash_records(db_file);
		}

		fclose(db_file);
		fscanf(db_index, "%s\n", buf1);
	}

	/* Exit if 0 records, unless this is shops */
	if (!rec_count) {
		if (mode == DB_BOOT_SHP || mode == DB_BOOT_GLD)
			return;
		log("SYSERR: boot error - 0 records counted in %s/%s.", prefix,
			 index_filename);
		exit(1);
	}

	/*
	 * NOTE: "bytes" does _not_ include strings or other later malloc'd things.
	 */
	switch (mode) {
		case DB_BOOT_WLD:
			CREATE(world, struct room_data, rec_count);
			size[0] = sizeof(struct room_data) * rec_count;
			log("   %d rooms, %d bytes.", rec_count, size[0]);
			break;
		case DB_BOOT_MOB:
			CREATE(mob_proto, struct char_data, rec_count);
			CREATE(mob_index, struct index_data, rec_count);
			size[0] = sizeof(struct index_data) * rec_count;
			size[1] = sizeof(struct char_data) * rec_count;
			log("   %d mobs, %d bytes in index, %d bytes in prototypes.", rec_count, size[0], size[1]);
			break;
		case DB_BOOT_OBJ:
			CREATE(obj_proto, struct obj_data, rec_count);
			CREATE(obj_index, struct index_data, rec_count);
			size[0] = sizeof(struct index_data) * rec_count;
			size[1] = sizeof(struct obj_data) * rec_count;
			log("   %d objs, %d bytes in index, %d bytes in prototypes.", rec_count, size[0], size[1]);
			break;
		case DB_BOOT_ZON:
			CREATE(zone_table, struct zone_data, rec_count);
			size[0] = sizeof(struct zone_data) * rec_count;
			log("   %d zones, %d bytes.", rec_count, size[0]);
			break;
		case DB_BOOT_HLP:
			CREATE(help_table, struct help_index_element, rec_count);
			size[0] = sizeof(struct help_index_element) * rec_count;
			log("   %d entries, %d bytes.", rec_count, size[0]);
			break;
	}

	rewind(db_index);
	fscanf(db_index, "%s\n", buf1);
	while (*buf1 != '$') {
		snprintf(buf2, sizeof(buf2), "%s%s", prefix, buf1);
		if (!(db_file = fopen(buf2, "r"))) {
			log("SYSERR: %s: %s", buf2, strerror(errno));
			exit(1);
		}
		switch (mode) {
			case DB_BOOT_WLD:
			case DB_BOOT_OBJ:
			case DB_BOOT_MOB:
				discrete_load(db_file, mode, buf2);
				break;
			case DB_BOOT_ZON:
				load_zones(db_file, buf2);
				break;
			case DB_BOOT_HLP:
				/*
				 * If you think about it, we have a race here.  Although, this is the
				 * "point-the-gun-at-your-own-foot" type of race.
				 */
				load_help(db_file);
				break;
			case DB_BOOT_SHP:
				boot_the_shops(db_file, buf2, rec_count);
				break;
			case DB_BOOT_GLD:
				boot_the_guilds(db_file, buf2, rec_count);
				break;
		}

		fclose(db_file);
		fscanf(db_index, "%s\n", buf1);
	}
	fclose(db_index);

	/* sort the help index */
	if (mode == DB_BOOT_HLP) {
		qsort(help_table, top_of_helpt, sizeof(struct help_index_element), hsort);
		top_of_helpt--;
	}
}


void discrete_load(FILE *fl, int mode, char *filename)
{
	int nr = -1, last;
	char line[READ_SIZE];

	const char *modes[] = {"world", "mob", "obj"};

	for (;;) {
		/*
		 * we have to do special processing with the obj files because they have
		 * no end-of-record marker :(
		 */
		if (mode != DB_BOOT_OBJ || nr < 0)
			if (!get_line(fl, line)) {
				if (nr == -1) {
					log("SYSERR: %s file %s is empty!", modes[mode], filename);
				} else {
					log("SYSERR: Format error in %s after %s #%d\n"
						 "...expecting a new %s, but file ended!\n"
						 "(maybe the file is not terminated with '$'?)", filename,
						 modes[mode], nr, modes[mode]);
				}
				exit(1);
			}
		if (*line == '$')
			return;

		if (*line == '#') {
			last = nr;
			if (sscanf(line, "#%d", &nr) != 1) {
				log("SYSERR: Format error after %s #%d", modes[mode], last);
				exit(1);
			}
			if (nr >= 99999)
				return;
			else
				switch (mode) {
					case DB_BOOT_WLD:
						parse_room(fl, nr);
						break;
					case DB_BOOT_MOB:
						parse_mobile(fl, nr);
						break;
					case DB_BOOT_OBJ:
						strlcpy(line, parse_object(fl, nr), sizeof(line));
						break;
				}
		} else {
			log("SYSERR: Format error in %s file %s near %s #%d", modes[mode],
				 filename, modes[mode], nr);
			log("SYSERR: ... offending line: '%s'", line);
			exit(1);
		}
	}
}

bitvector_t asciiflag_conv(char *flag)
{
	bitvector_t flags = 0;
	int is_num = TRUE;
	char *p;

	for (p = flag; *p; p++) {
		if (islower(*p))
			flags |= 1 << (*p - 'a');
		else if (isupper(*p))
			flags |= 1 << (26 + (*p - 'A'));

		if (!isdigit(*p))
			is_num = FALSE;
	}

	if (is_num)
		flags = atol(flag);

	return (flags);
}


void get_one_line(FILE *fl, char *buf)
{
	if (fgets(buf, READ_SIZE, fl) == NULL) {
		log("SYSERR: error reading help file: not terminated with $?");
		exit(1);
	}

	buf[strlen(buf) - 1] = '\0'; /* take off the trailing \n */
}


void free_help(void)
{
	int hp;

	if (!help_table)
		return;

	for (hp = 0; hp <= top_of_helpt; hp++) {
		if (help_table[hp].keyword)
			free(help_table[hp].keyword);
		if (help_table[hp].entry && !help_table[hp].duplicate)
			free(help_table[hp].entry);
	}

	free(help_table);
	help_table = NULL;
	top_of_helpt = 0;
}


void load_help(FILE *fl)
{
#if defined(CIRCLE_MACINTOSH)
	static char key[READ_SIZE + 1], next_key[READ_SIZE + 1], entry[32384]; /* too big for stack? */
#else
	char key[READ_SIZE + 1], next_key[READ_SIZE + 1], entry[32384];
#endif
	size_t entrylen;
	char line[READ_SIZE + 1], *scan;
	struct help_index_element el;

	/* get the first keyword line */
	get_one_line(fl, key);
	while (*key != '$') {
		strcat(key, "\r\n");	/* strcat: OK (READ_SIZE - "\n" + "\r\n" == READ_SIZE + 1) */
		entrylen = strlcpy(entry, key, sizeof(entry));

		/* read in the corresponding help entry */
		get_one_line(fl, line);
		while (*line != '#' && entrylen < sizeof(entry) - 1) {
			entrylen += strlcpy(entry + entrylen, line, sizeof(entry) - entrylen);

			if (entrylen + 2 < sizeof(entry) - 1) {
				strcpy(entry + entrylen, "\r\n");	/* strcpy: OK (size checked above) */
				entrylen += 2;
			}
			get_one_line(fl, line);
		}

		if (entrylen >= sizeof(entry) - 1) {
			int keysize;
			const char *truncmsg = "\r\n*TRUNCATED*\r\n";

			strcpy(entry + sizeof(entry) - strlen(truncmsg) - 1, truncmsg);	/* strcpy: OK (assuming sane 'entry' size) */

			keysize = strlen(key) - 2;
			log("SYSERR: Help entry exceeded buffer space: %.*s", keysize, key);

			/* If we ran out of buffer space, eat the rest of the entry. */
			while (*line != '#')
				get_one_line(fl, line);
		}

		/* now, add the entry to the index with each keyword on the keyword line */
		el.duplicate = 0;
		el.entry = strdup(entry);
		scan = one_word(key, next_key);
		while (*next_key) {
			el.keyword = strdup(next_key);
			help_table[top_of_helpt++] = el;
			el.duplicate++;
			scan = one_word(scan, next_key);
		}

		/* get next keyword line (or $) */
		get_one_line(fl, key);
	}
}


int hsort(const void *a, const void *b)
{
	const struct help_index_element *a1, *b1;

	a1 = (const struct help_index_element *) a;
	b1 = (const struct help_index_element *) b;

	return (str_cmp(a1->keyword, b1->keyword));
}


/************************************************************************
 *  funcs of a (more or less) general utility nature			*
 ************************************************************************/

char fread_letter(FILE *fp)
{
   char c;

   do {
		c = getc(fp);
   } while (isspace(c));

   return c;
}
 

/* read and allocate space for a '~'-terminated string from a given file */
char *fread_string(FILE *fl, const char *error)
{
	char buf[MAX_STRING_LENGTH], tmp[512];
	char *point;
	int done = 0, length = 0, templength;

	*buf = '\0';

	do {
		if (!fgets(tmp, 512, fl)) {
			log("SYSERR: fread_string: format error at or near %s", error);
			exit(1);
		}
		/* If there is a '~', end the string; else put an "\r\n" over the '\n'. */
		if ((point = strchr(tmp, '~')) != NULL) {
			*point = '\0';
			done = 1;
		} else {
			point = tmp + strlen(tmp) - 1;
			*(point++) = '\r';
			*(point++) = '\n';
			*point = '\0';
		}

		templength = strlen(tmp);

		if (length + templength >= MAX_STRING_LENGTH) {
			log("SYSERR: fread_string: string too large (db.c)");
			log("%s", error);
			exit(1);
		} else {
			strcat(buf + length, tmp);	/* strcat: OK (size checked above) */
			length += templength;
		}
	} while (!done);

	/* allocate space for the new string and copy it */
	return (strlen(buf) ? strdup(buf) : NULL);
}


/*
 * Steps:
 *   1: Read contents of a text file.
 *   2: Make sure no one is using the pointer in paging.
 *   3: Allocate space.
 *   4: Point 'buf' to it.
 *
 * We don't want to free() the string that someone may be
 * viewing in the pager.  page_string() keeps the internal
 * strdup()'d copy on ->showstr_head and it won't care
 * if we delete the original.  Otherwise, strings are kept
 * on ->showstr_vector but we'll only match if the pointer
 * is to the string we're interested in and not a copy.
 *
 * If someone is reading a global copy we're trying to
 * replace, give everybody using it a different copy so
 * as to avoid special cases.
 */
int file_to_string_alloc(const char *name, char **buf)
{
	int temppage;
	char temp[MAX_STRING_LENGTH];
	struct descriptor_data *in_use;

	for (in_use = descriptor_list; in_use; in_use = in_use->next)
		if (in_use->showstr_vector && *in_use->showstr_vector == *buf)
			return (-1);

	/* Lets not free() what used to be there unless we succeeded. */
	if (file_to_string(name, temp) < 0)
		return (-1);

	for (in_use = descriptor_list; in_use; in_use = in_use->next) {
		if (!in_use->showstr_count || *in_use->showstr_vector != *buf)
			continue;

		/* Let's be nice and leave them at the page they were on. */
		temppage = in_use->showstr_page;
		paginate_string((in_use->showstr_head = strdup(*in_use->showstr_vector)), in_use);
		in_use->showstr_page = temppage;
	}

	if (*buf)
		free(*buf);

	*buf = strdup(temp);
	return (0);
}


/* read contents of a text file, and place in buf */
int file_to_string(const char *name, char *buf)
{
	FILE *fl;
	char tmp[READ_SIZE + 3];
	int len;

	*buf = '\0';

	if (!(fl = fopen(name, "r"))) {
		log("SYSERR: reading %s: %s", name, strerror(errno));
		return (-1);
	}

	for (;;) {
		if (!fgets(tmp, READ_SIZE, fl))	/* EOF check */
			break;
		if ((len = strlen(tmp)) > 0)
			tmp[len - 1] = '\0'; /* take off the trailing \n */
		strcat(tmp, "\r\n");	/* strcat: OK (tmp:READ_SIZE+3) */

		if (strlen(buf) + strlen(tmp) + 1 > MAX_STRING_LENGTH) {
			log("SYSERR: %s: string too big (%d max)", name, MAX_STRING_LENGTH);
			*buf = '\0';
			fclose(fl);
			return (-1);
		}
		strcat(buf, tmp);	/* strcat: OK (size checked above) */
	}

	fclose(fl);

	return (0);
}


int check_bitvector_names(bitvector_t bits, size_t namecount, const char *whatami, const char *whatbits)
{
	unsigned int flagnum;
	bool error = FALSE;

	/* See if any bits are set above the ones we know about. */
	if (bits <= (~(bitvector_t)0 >> (sizeof(bitvector_t) * 8 - namecount)))
		return (FALSE);

	for (flagnum = namecount; flagnum < sizeof(bitvector_t) * 8; flagnum++)
		if ((1 << flagnum) & bits) {
			log("SYSERR: %s has unknown %s flag, bit %d (0 through %d known).",
				 whatami, whatbits, flagnum, namecount - 1);
			error = TRUE;
		}

	return (error);
}
