#include "conf.h"
#include "sysdep.h"

#include "structs.h"
#include "utils.h"

#include "comm.h"
#include "constants.h"
#include "db.h"
#include "handler.h"
#include "house.h"
#include "interpreter.h"


/* extern vars */
extern int mini_mud;
extern struct reset_q_type reset_q;
extern struct time_info_data time_info;

/* local functions */
void renum_zone_table(void);
void load_zones(FILE *fl, char *zonename);
void zone_update(void);
void log_zone_error(zone_rnum zone, int cmd_no, const char *message);
void reset_zone(zone_rnum zone);
room_rnum real_zone(room_vnum vnum);


typedef const char *(*verify_func)(int);

#define ZCMD zone_table[zone].cmd[cmd_no]

/*
 * "resulve vnums into rnums in the zone reset tables"
 *
 * Or in English: Once all of the zone reset tables have been loaded, we
 * resolve the virtual numbers into real numbers all at once so we don't have
 * to do it repeatedly while the game is running.  This does make adding any
 * room, mobile, or object a little more difficult while the game is running.
 *
 * NOTE 1: Assumes NOWHERE == NOBODY == NOTHING.
 * NOTE 2: Assumes sizeof(room_rnum) >= (sizeof(mob_rnum) and sizeof(obj_rnum))
 */
void renum_zone_table(void)
{
	int cmd_no;
	room_rnum a, b, c, olda, oldb, oldc;
	zone_rnum zone;
	char buf[128];

	for (zone = 0; zone <= top_of_zone_table; zone++)
		for (cmd_no = 0; ZCMD.command != 'S'; cmd_no++) {
			a = b = c = 0;
			olda = ZCMD.arg1;
			oldb = ZCMD.arg2;
			oldc = ZCMD.arg3;
			switch (ZCMD.command) {
				case 'M':
					a = ZCMD.arg1 = real_mobile(ZCMD.arg1);
					c = ZCMD.arg3 = real_room(ZCMD.arg3);
					break;
				case 'O':
					a = ZCMD.arg1 = real_object(ZCMD.arg1);
					if (ZCMD.arg3 != NOWHERE)
						c = ZCMD.arg3 = real_room(ZCMD.arg3);
					break;
				case 'G':
					a = ZCMD.arg1 = real_object(ZCMD.arg1);
					break;
				case 'E':
					a = ZCMD.arg1 = real_object(ZCMD.arg1);
					break;
				case 'P':
					a = ZCMD.arg1 = real_object(ZCMD.arg1);
					c = ZCMD.arg3 = real_object(ZCMD.arg3);
					break;
				case 'D':
					a = ZCMD.arg1 = real_room(ZCMD.arg1);
					break;
				case 'R': /* rem obj from room */
					a = ZCMD.arg1 = real_room(ZCMD.arg1);
					b = ZCMD.arg2 = real_object(ZCMD.arg2);
					break;
			}
			if (a == NOWHERE || b == NOWHERE || c == NOWHERE) {
				if (!mini_mud) {
					snprintf(buf, sizeof(buf), "Invalid vnum %d, cmd disabled",
								a == NOWHERE ? olda : b == NOWHERE ? oldb : oldc);
					log_zone_error(zone, cmd_no, buf);
				}
				ZCMD.command = '*';
			}
		}
}



#define Z	zone_table[zone]

/* load the zone table and command tables */
void load_zones(FILE *fl, char *zonename)
{
	static zone_rnum zone = 0;
	int cmd_no, num_of_cmds = 0, line_num = 0, tmp, error;
	char *ptr, buf[READ_SIZE], zname[READ_SIZE], buf2[MAX_STRING_LENGTH];

	strlcpy(zname, zonename, sizeof(zname));

	/* Skip first 3 lines lest we mistake the zone name for a command. */
	for (tmp = 0; tmp < 3; tmp++)
		get_line(fl, buf);

	/*  More accurate count. Previous was always 4 or 5 too high. -gg 2001/1/17
	 *  Note that if a new zone command is added to reset_zone(), this string
	 *  will need to be updated to suit. - ae.
	 */
	while (get_line(fl, buf)) {
		if ((strchr("MOPGERD", buf[0]) && buf[1] == ' ') ||
			 (buf[0] == 'S' && buf[1] == '\0'))
			num_of_cmds++;
	}

	rewind(fl);

	if (num_of_cmds == 0) {
		log("SYSERR: %s is empty!", zname);
		exit(1);
	} else
		CREATE(Z.cmd, struct reset_com, num_of_cmds);

	line_num += get_line(fl, buf);

	if (sscanf(buf, "#%hd", &Z.number) != 1) {
		log("SYSERR: Format error in %s, line %d", zname, line_num);
		exit(1);
	}
	snprintf(buf2, sizeof(buf2), "beginning of zone #%d", Z.number);

	line_num += get_line(fl, buf);
	if ((ptr = strchr(buf, '~')) != NULL)	/* take off the '~' if it's there */
		*ptr = '\0';
	Z.name = strdup(buf);

	line_num += get_line(fl, buf);
	if (sscanf(buf, " %hd %hd %d %d %d ", &Z.bot, &Z.top, &Z.lifespan, &Z.reset_mode, &Z.zflags) != 5) {
		log("SYSERR: Format error in numeric constant line of %s", zname);
		exit(1);
	}
	if (Z.bot > Z.top) {
		log("SYSERR: Zone %d bottom (%d) > top (%d).", Z.number, Z.bot, Z.top);
		exit(1);
	}

	Z.pressure = 960;
	if ((time_info.month >= 7) && (time_info.month <= 12))
		Z.pressure += dice(1, 50);
	else
		Z.pressure += dice(1, 80);
	Z.change = 0;
	if (Z.pressure <= 980)
		Z.sky = SKY_LIGHTNING;
	else if (Z.pressure <= 1000)
		Z.sky = SKY_RAINING;
	else if (Z.pressure <= 1020)
		Z.sky = SKY_CLOUDY;
	else
		Z.sky = SKY_CLOUDLESS;

   line_num += get_line(fl, buf);
   if (sscanf(buf, " %d %d %d ", &Z.lvl1, &Z.lvl2, &Z.status_mode) != 3) {
		log("SYSERR: Format error in 3-constant line of %s", zname);
		exit(1);
   }
	if (Z.lvl1 > Z.lvl2 || Z.lvl1 < 1 || Z.lvl2 > LVL_IMPL) 
		log("SYSERR: Invalid level range for zone %d (%s)", Z.number, zname);

	cmd_no = 0;

	for (;;) {
		if ((tmp = get_line(fl, buf)) == 0) {
			log("SYSERR: Format error in %s - premature end of file", zname);
			exit(1);
		}
		line_num += tmp;
		ptr = buf;
		skip_spaces(&ptr);

		if ((ZCMD.command = *ptr) == '*')
			continue;

		ptr++;

		if (ZCMD.command == 'S' || ZCMD.command == '$') {
			ZCMD.command = 'S';
			break;
		}
		error = 0;
		if (strchr("MOEPD", ZCMD.command) == NULL) {	/* a 3-arg command */
			if (sscanf(ptr, " %d %d %d ", &tmp, &ZCMD.arg1, &ZCMD.arg2) != 3)
				error = 1;
		} else {
			if (sscanf(ptr, " %d %d %d %d ", &tmp, &ZCMD.arg1, &ZCMD.arg2,
						  &ZCMD.arg3) != 4)
				error = 1;
		}

		ZCMD.if_flag = tmp;

		if (error) {
			log("SYSERR: Format error in %s, line %d: '%s'", zname, line_num, buf);
			exit(1);
		}
		ZCMD.line = line_num;
		cmd_no++;
	}

	if (num_of_cmds != cmd_no + 1) {
		log("SYSERR: Zone command count mismatch for %s. Estimated: %d, Actual: %d", zname, num_of_cmds, cmd_no + 1);
		exit(1);
	}

	top_of_zone_table = zone++;
}

#undef Z


#define ZO_DEAD  999

/* update zone ages, queue for reset if necessary, and dequeue when possible */
void zone_update(void)
{
	int i;
	struct reset_q_element *update_u, *temp;
	static int timer = 0;

	/* jelson 10/22/92 */
	if (((++timer * PULSE_ZONE) / PASSES_PER_SEC) >= 60) {
		/* one minute has passed */
		/*
		 * NOT accurate unless PULSE_ZONE is a multiple of PASSES_PER_SEC or a
		 * factor of 60
		 */

		timer = 0;

		/* since one minute has passed, increment zone ages */
		for (i = 0; i <= top_of_zone_table; i++) {
			if (zone_table[i].age < zone_table[i].lifespan &&
				 zone_table[i].reset_mode)
				(zone_table[i].age)++;

			if (zone_table[i].age >= zone_table[i].lifespan &&
				 zone_table[i].age < ZO_DEAD && zone_table[i].reset_mode) {
				/* enqueue zone */

				CREATE(update_u, struct reset_q_element, 1);

				update_u->zone_to_reset = i;
				update_u->next = 0;

				if (!reset_q.head)
					reset_q.head = reset_q.tail = update_u;
				else {
					reset_q.tail->next = update_u;
					reset_q.tail = update_u;
				}

				zone_table[i].age = ZO_DEAD;
			}
		}
	}	/* end - one minute has passed */


	/* dequeue zones (if possible) and reset */
	/* this code is executed every 10 seconds (i.e. PULSE_ZONE) */
	for (update_u = reset_q.head; update_u; update_u = update_u->next)
		if (zone_table[update_u->zone_to_reset].reset_mode == 2 ||
			 is_empty(update_u->zone_to_reset)) {
			reset_zone(update_u->zone_to_reset);
			mudlog(CMP, LVL_GOD, FALSE, "Auto zone reset: %s", zone_table[update_u->zone_to_reset].name);
			/* dequeue */
			if (update_u == reset_q.head)
				reset_q.head = reset_q.head->next;
			else {
				for (temp = reset_q.head; temp->next != update_u;
					  temp = temp->next);

				if (!update_u->next)
					reset_q.tail = temp;

				temp->next = update_u->next;
			}

			free(update_u);
			break;
		}
}

void log_zone_error(zone_rnum zone, int cmd_no, const char *message)
{
	mudlog(NRM, LVL_GOD, TRUE, "SYSERR: zone file: %s", message);
	mudlog(NRM, LVL_GOD, TRUE, "SYSERR: ...offending cmd: '%c' cmd in zone #%d, line %d",
			 ZCMD.command, zone_table[zone].number, ZCMD.line);
}

#define ZONE_ERROR(message) \
	{ log_zone_error(zone, cmd_no, message); last_cmd = 0; }

/* execute the reset command table of a given zone */
void reset_zone(zone_rnum zone)
{
	int cmd_no, last_cmd = 0;
	struct char_data *mob = NULL;
	struct obj_data *obj, *obj_to;

	for (cmd_no = 0; ZCMD.command != 'S'; cmd_no++) {

		if (ZCMD.if_flag && !last_cmd)
			continue;

		/*  This is the list of actual zone commands.  If any new
		 *  zone commands are added to the game, be certain to update
		 *  the list of commands in load_zone() so that the counting
		 *  will still be correct. - ae.
		 */
		switch (ZCMD.command) {
			case '*':			/* ignore command */
				last_cmd = 0;
				break;

			case 'M':			/* read a mobile */
				if (mob_index[ZCMD.arg1].number < ZCMD.arg2) {
					mob = read_mobile(ZCMD.arg1, REAL);
					char_to_room(mob, ZCMD.arg3);
					last_cmd = 1;
				} else
					last_cmd = 0;
				break;

			case 'O':			/* read an object */
				if (obj_index[ZCMD.arg1].number < ZCMD.arg2) {
					if (ZCMD.arg3 != NOWHERE) {
						obj = read_object(ZCMD.arg1, REAL);
						obj_to_room(obj, ZCMD.arg3);
						last_cmd = 1;
					} else {
						obj = read_object(ZCMD.arg1, REAL);
						IN_ROOM(obj) = NOWHERE;
						last_cmd = 1;
					}
				} else
					last_cmd = 0;
				break;

			case 'P':			/* object to object */
				if (obj_index[ZCMD.arg1].number < ZCMD.arg2) {
					obj = read_object(ZCMD.arg1, REAL);
					if (!(obj_to = get_obj_num(ZCMD.arg3))) {
						ZONE_ERROR("target obj not found, command disabled");
						ZCMD.command = '*';
						break;
					}
					obj_to_obj(obj, obj_to);
					last_cmd = 1;
				} else
					last_cmd = 0;
				break;

			case 'G':			/* obj_to_char */
				if (!mob) {
					ZONE_ERROR("attempt to give obj to non-existant mob, command disabled");
					ZCMD.command = '*';
					break;
				}
				if (obj_index[ZCMD.arg1].number < ZCMD.arg2) {
					obj = read_object(ZCMD.arg1, REAL);
					obj_to_char(obj, mob);
					last_cmd = 1;
				} else
					last_cmd = 0;
				break;

			case 'E':			/* object to equipment list */
				if (!mob) {
					ZONE_ERROR("trying to equip non-existant mob, command disabled");
					ZCMD.command = '*';
					break;
				}
				if (obj_index[ZCMD.arg1].number < ZCMD.arg2) {
					if (ZCMD.arg3 < 0 || ZCMD.arg3 >= NUM_WEARS) {
						ZONE_ERROR("invalid equipment pos number");
					} else {
						obj = read_object(ZCMD.arg1, REAL);
						equip_char(mob, obj, ZCMD.arg3);
						last_cmd = 1;
					}
				} else
					last_cmd = 0;
				break;

			case 'R': /* rem obj from room */
				if ((obj = get_obj_in_list_num(ZCMD.arg2, world[ZCMD.arg1].contents)) != NULL)
					extract_obj(obj);
				last_cmd = 1;
				break;


			case 'D':			/* set state of door */
				if (ZCMD.arg2 < 0 || ZCMD.arg2 >= NUM_OF_DIRS ||
					 (world[ZCMD.arg1].dir_option[ZCMD.arg2] == NULL)) {
					ZONE_ERROR("door does not exist, command disabled");
					ZCMD.command = '*';
				} else
					switch (ZCMD.arg3) {
						case 0:
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_CLOSED = 0;
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_LOCKED = 0;
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_HIDDEN = 0;
							break;
						case 1:
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_CLOSED = 1;
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_LOCKED = 0;
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_HIDDEN = 0;
							break;
						case 2:
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_CLOSED = 1;
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_LOCKED = 1;
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_HIDDEN = 0;
							break;
						case 3:
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_CLOSED = 0;
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_LOCKED = 0;
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_HIDDEN = 1;
							break;
						case 4:
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_CLOSED = 1;
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_LOCKED = 0;
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_HIDDEN = 1;
							break;
						case 5:
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_CLOSED = 1;
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_LOCKED = 1;
							world[ZCMD.arg1].dir_option[ZCMD.arg2]->exit_info.EX_HIDDEN = 1;
							break;
					}
				last_cmd = 1;
				break;

			default:
				ZONE_ERROR("unknown cmd in reset table; cmd disabled");
				ZCMD.command = '*';
				break;
		}
	}

	zone_table[zone].age = 0;
}



/* for use in reset_zone; return TRUE if zone 'nr' is free of PC's  */
int is_empty(zone_rnum zone_nr)
{
	struct descriptor_data *i;

	for (i = descriptor_list; i; i = i->next) {
		if (STATE(i) != CON_PLAYING)
			continue;
		if (IN_ROOM(i->character) == NOWHERE)
			continue;
		if (GET_LEVEL(i->character) >= LVL_IMMORT)
			continue;
		if (world[IN_ROOM(i->character)].zone != zone_nr)
			continue;

		return (0);
	}

	return (1);
}



/* returns the real number of the zone with given virtual number */
room_rnum real_zone(room_vnum vnum)
{
	room_rnum bot, top, mid;

	bot = 0;
	top = top_of_zone_table;

	/* perform binary search on zone-table */
	for (;;) {
		mid = (bot + top) / 2;

		if ((zone_table + mid)->number == vnum)
			return (mid);
		if (bot >= top)
			return (NOWHERE);
		if ((zone_table + mid)->number > vnum)
			top = mid - 1;
		else
			bot = mid + 1;
	}
}


#define VALID_EXIT(x,y) (world[(x)].dir_option[(y)] && TOROOM(x, y) != NOWHERE)
#define IS_DOOR(x,y)    (EXIT_FLAGGED(world[(x)].dir_option[(y)], EX_ISDOOR))

const char *check_exits(int rnum)
{
	static char retbuf[MAX_STRING_LENGTH];
	int i;
	size_t nlen = 0, len = 0;

	if (rnum == NOWHERE)
		return NULL;

	retbuf[0] = 0;

	for (i = 0; i < NUM_OF_DIRS; i++) {
		if (VALID_EXIT(rnum, i)) {
			if (H_TOROOM(H_TOROOM(rnum, i), rev_dir[i]) != rnum) {
				nlen = snprintf(retbuf + len, sizeof(retbuf) - len,
									 "Room %d: going %s then %s doesn't return to "
									 "starting room\r\n", world[rnum].number, 
									 dirs[i], dirs[rev_dir[i]]);

				if (len + nlen >= sizeof(retbuf) || nlen < 0)
					break;
				len += nlen;
			}
			if (H_TOROOM(rnum, i) == NOWHERE) {
				nlen = snprintf(retbuf + len, sizeof(retbuf) - len,
									 "Room %d: door to the %s leads nowhere\r\n",
									 world[rnum].number, dirs[i]);

				if (len + nlen >= sizeof(retbuf) || nlen < 0)
					break;
				len += nlen;
			}
			if (IS_DOOR(rnum, i) && 
				 (!VALID_EXIT(H_TOROOM(rnum, i), rev_dir[i]) || 
				  !IS_DOOR(H_TOROOM(rnum, i), rev_dir[i]))) {
				nlen = snprintf(retbuf + len, sizeof(retbuf) - len,
									 "Room %d: door to the %s doesn't have a door on "
									 "the other side\r\n", world[rnum].number, dirs[i]);
				if (len + nlen >= sizeof(retbuf) || nlen < 0)
					break;
				len += nlen;

			}
			if (IS_CLOSED(rnum, i) && 
				 (!VALID_EXIT(H_TOROOM(rnum, i), rev_dir[i]) || 
				  !IS_CLOSED(H_TOROOM(rnum, i), rev_dir[i]))) {
				nlen = snprintf(retbuf + len, sizeof(retbuf) - len, 
									 "Room %d: door to the %s is closed, but other "
									 "side isn't\r\n", world[rnum].number, dirs[i]);

				if (len + nlen >= sizeof(retbuf) || nlen < 0)
					break;
				len += nlen;
			}
		}
	}

	return (*retbuf ? retbuf : NULL);
}


ACMD(do_verify)
{
	int a = -1, b = -1, i, vfn;
	verify_func vfs[] = { check_exits, NULL };
	const char *s;
	char buf[MAX_STRING_LENGTH];

	sscanf(argument, "%d %d", &a, &b);

	buf[0] = 0;

	if (a < 0 || b < 0 || a == b) {
		send_to_char(ch, "Usage: verify <first room> <last room>\r\n");
		return;
	}

	for (vfn = 0; vfs[vfn]; vfn++) {
		for (i = a; i <= b; i++) {
			s = vfs[vfn](real_room(i));
			if (s)
				strcat(buf, s);
		}
	}

	if (buf[0])
		page_string(ch->desc, buf, TRUE);
	else
		send_to_char(ch, "No errors found.\r\n");
}

