/* ************************************************************************
 *   File: fight.c                                       Part of CircleMUD *
 *  Usage: Combat system                                                   *
 *                                                                         *
 *  All rights reserved.  See license.doc for complete information.        *
 *                                                                         *
 *  Copyright (C) 1993, 94 by the Trustees of the Johns Hopkins University *
 *  CircleMUD is based on DikuMUD, Copyright (C) 1990, 1991.               *
 ************************************************************************ */

#include "conf.h"
#include "sysdep.h"

#include "structs.h"
#include "utils.h"

#include "coins.h"
#include "comm.h"
#include "constants.h"
#include "db.h"
#include "handler.h"
#include "interpreter.h"
#include "mobprog.h"
#include "screen.h"
#include "spells.h"

/* Structures */
struct char_data *combat_list = NULL;	/* head of l-list of fighting chars */
struct char_data *next_combat_list = NULL;

/* External structures */
extern struct message_list fight_messages[MAX_MESSAGES];
extern int pk_allowed;		/* see config.c */
extern int max_exp_gain;	/* see config.c */
extern int max_exp_loss;	/* see config.c */
extern int max_npc_corpse_time, max_pc_corpse_time;

/* External procedures */
ACMD(do_flee);
ACMD(do_get);
ACMD(do_sac);
ACMD(do_split);

SPECIAL(blind_weapon);
SPECIAL(curse_weapon);
SPECIAL(fireball_weapon);

char *fread_action(FILE *fl, int nr);
int backstab_mult(int level);
int thaco(int ch_class, int level);
int ok_damage_shopkeeper(struct char_data *ch, struct char_data *victim);

/* local functions */
void perform_group_gain(struct char_data *ch, int base, struct char_data *victim);
void dam_message(int dam, struct char_data *ch, struct char_data *victim, int w_type);
void appear(struct char_data *ch);
void load_messages(void);
void free_messages(void);
void free_messages_type(struct msg_type *msg);
void check_killer(struct char_data *ch, struct char_data *vict);
void make_corpse(struct char_data *ch);
void change_alignment(struct char_data *ch, struct char_data *victim);
void death_cry(struct char_data *ch);
void raw_kill(struct char_data *ch, struct char_data *killer);
void die(struct char_data *ch, struct char_data *killer);
void group_gain(struct char_data *ch, struct char_data *victim);
void solo_gain(struct char_data *ch, struct char_data *victim);
char *replace_string(const char *str, const char *weapon_singular, const char *weapon_plural);
void perform_violence(void);
int compute_armor_class(struct char_data *ch);
int compute_thaco(struct char_data *ch, struct char_data *vict);
void double_hit(struct char_data * ch, struct char_data * victim, int type);



#define NUM_OF_FUN_PARTS 21
struct fun_body_piece {
   int number;       /* this parts number */
   char name[40];    /* names of this part */
   int  nname;       /* some parts you couldn't trace to an owner*/
   char sdesc[128];  /* short desc: that of inventory  */
   char rdesc[128];  /* room desc: when on ground */
   int  take;        /* some body parts don't transfer well */
   char actout[128]; /* what people in room see upon death, using	act()*/
   char actkil[128]; /* what the killer sees upon dismemberment, using act() */
};

struct fun_body_piece parts[NUM_OF_FUN_PARTS] = {
	{0, "eyeball eye", 
	 1, "the eye of %s", "The eyeball of %s is lying here.",
	 1, "$n's attack knocks an eye out of $N!",
	 "Your attack knocks an eye out of $N!"},
	{1, "liver", 
	 1, "the liver of %s", "%s's liver is lying here.",
	 1, "$n's cruel attack blows $N's liver out!",
	 "Your cruel attack blows $N's liver out!"},
	{2, "arm",
	 1, "one of %s's arms", "%s's arm is lying here on the ground.",
	 1, "$n's removes $N's arm!",
	 "You remove $N's arm!"},
	{3, "bowels",
	 1, "%s's bowels", "Ick. %s's bowels are lying here.",
	 1, "$n debowels $N!",
	 "You debowl $N!"},
	{4, "tush butt rear ass",
	 1, "%s's rear end", "Some one cut off %s's butt and left it here.",
	 1, "$n laughs as he severs $N's rear end!",
	 "You laugh as you sever $N's rear end!"},
	{5, "right leg",
	 1, "%s's right leg", "%s's right leg is here.",
	 1, "$N gracefully cuts his leg off! $n chortles merrily!",
	 "You watch in awe as $N cuts his leg off!"},
	{6, "left leg",
	 1, "the left leg of %s", "The left leg of %s is lying here.",
	 1, "$n's screams and strikes $N leg off at the hip!",
	 "With a scream of rage, you strike $N's leg off!"},
	{7, "head",
	 1, "%s's ugly head", "%s's head is lying here, staring at you.",
	 1, "$n severs $N's in a move composed of speed and grace!",
	 "With speed and grace, you sever $N's head!"},
	{8, "thumb",
	 1, "%s's thumb", "One of %s's thumbs is lying here.",
	 1, "$n's attack severs a thumb from $N!",
	 "Your attack severs a thumb from $N!"},
	{9, "finger",
	 1, "%s's finger", "One of %s fingers is lying here.",
	 1, "$n's attack severs a finger from $N!",
	 "Your attack severs a finger from $N!"},
	{10, "stomach",
	 1, "%s's stomach", "%s lost his stomach here.",
	 1, "With animal force, $n tears $N's stomach out!",
	 "With animal force, you tear $N's stomach out!"},
	{11, "heart",
	 1, "the once beating heart of %s", "%s's once beating heart lies here.",
	 1, "$n's uses pure strength to eviscirate $N!",
	 "Your depend on your fierce strength, and eviscerate $N!"},
	{12, "spine",
	 1, "the spine of %s", "The spine of %s is lying here.",
	 1, "$n's attack shatters $N's spine!",
	 "Your attack shatters $N's spine!"},
	{13, "intestine",
	 0, "An icky pile of intestines", "An icky pile of intestines is here - colon and all.",
	 0, "$n hits so hard, that $N pukes up his intestines !",
	 "You hit $N so hard that he pukes up his intestines!"},
	{14, "puke vomit",
	 0, "chunky vomit", "Some one upchucked on the floor here.",
	 0, "$N throws up all over!",
	 "$N throws up all over you!"},
	{15, "pool blood",
	 0, "A pool of blood", "Blood has formed a pool on the ground.",
	 0, "$N bleeds horrendously!",
	 "$N bleeds horrendously!"},
	{16, "riblet",
	 1,"a meaty %s riblet", "A meaty riblet from %s is lying here.",
	 1, "$n's explodes $N's chest with a barrage of attacks!",
	 "Your cause $N's chest to explode from a barrage of attacks!"},
	{17, "nose",
	 1, "%s's nose", "%s lost his nose here.",
	 1, "$n cackles gleefuly as he removes $N's nose!",
	 "You cackle as you sever $N's nose!"},
	{18, "ear",
	 1, "%s's ear", "%s's bloody severed ear is here.",
	 1, "$n's grabs $N's ear and rips it off!",
	 "Your rip off $N's ear!"},
	{19, "brain",
	 1, "the jiggly brain of %s", "The squishy brain of %s is here.",
	 1, "$n shatters $N's skull, knocking the brain on the ground!",
	 "You shatter $N's skull, knocking the brain on the ground!"},
	{20, "lung",
	 1, "a bloody lung from %s", "A blood soaked lung from %s.",
	 1, "$N screams his last as $n removes a lung!",
	 "$N's screams are cut short as you remove a lung!"}
};

void make_fun_body_pieces(struct char_data *ch, struct char_data *killer)
{
	struct obj_data *piece;
	char buf2[MAX_STRING_LENGTH];
	int i;

	/* Safety code.  -ejg */
	if (!ch || !killer)
		return;

	/* Llets check and see if we even GET body parts, eh?  I mean,
	 * they're fun, but it wouldn't be quite as fun if they were 
	 * always there!*/

	if (rand_number(1, 5) < 4)
		return;

	/* Then Horray! We's got parts! */
	/* But which part? */
	i = rand_number(0, 20);   /* 21 pieces should be okay*/
	piece = create_obj();

	/* Now, everything we have should be in the structures, neh? */

	/* Name first */
	piece->name = strdup(parts[i].name);

	/* Then lets see about the descs */
	if (parts[i].nname) {
		snprintf(buf2, sizeof(buf2), parts[i].sdesc, GET_NAME(ch));
		piece->short_description = strdup(buf2);

		snprintf(buf2, sizeof(buf2), parts[i].rdesc, GET_NAME(ch));
		piece->description = strdup(buf2);
	} else {
		piece->short_description = strdup(parts[i].sdesc);
		piece->description = strdup(parts[i].rdesc);
	}

	/* Well, now we know how it looks, lets see if we wanna take it. */
	if (parts[i].take)
		GET_OBJ_WEAR(piece).ITEM_WEAR_TAKE = 1;

	/* And lets see how it got here in the first place, neh? */
	act(parts[i].actout, FALSE, killer, 0, ch, TO_ROOM);
	act(parts[i].actkil, FALSE, killer, 0, ch, TO_CHAR);

	/* setup the rest of the stats any object needs */
	piece->item_number    = NOTHING;
	IN_ROOM(piece)        = NOWHERE;
	GET_OBJ_TYPE(piece)   = ITEM_CONTAINER;
	GET_OBJ_VAL(piece, 0) = 0;   /* You can't store stuff in a corpse */
	GET_OBJ_VAL(piece, 3) = 1;   /* corpse identifier */
	GET_OBJ_EXTRA(piece).ITEM_NORENT = 1;
	GET_OBJ_EXTRA(piece).ITEM_NOSELL = 1;
	GET_OBJ_EXTRA(piece).ITEM_NODONATE = 1;
	GET_OBJ_WEIGHT(piece) = 1;
	GET_OBJ_RENT(piece)   = 1;

	/* Note - you may have some trouble with corpse decay here -
	 *	improper settings WILL cause the mud to crash if you do not
	 *	correcly decay.  Right now, the pieces are setup as a corpse, so
	 *	if you made any changes to your corpse identifiers, fix it
	 *	above. */

	GET_OBJ_TIMER(piece) = max_npc_corpse_time;

	/* and thats all folks! */
	obj_to_room(piece, IN_ROOM(ch));
}


/* Weapon attack texts */
struct attack_hit_type attack_hit_text[NUM_ATTACK_TYPES] =
{
	{"hit", "hits"},		/* 0 */
	{"sting", "stings"},
	{"whip", "whips"},
	{"slash", "slashes"},
	{"bite", "bites"},
	{"bludgeon", "bludgeons"},	/* 5 */
	{"crush", "crushes"},
	{"pound", "pounds"},
	{"claw", "claws"},
	{"maul", "mauls"},
	{"thrash", "thrashes"},	/* 10 */
	{"pierce", "pierces"},
	{"blast", "blasts"},
	{"punch", "punches"},
	{"stab", "stabs"}
};

#define IS_WEAPON(type) (((type) >= TYPE_HIT) && ((type) < TYPE_SUFFERING))

/* The Fight related routines */

void appear(struct char_data *ch)
{
	if (affected_by_spell(ch, SPELL_INVISIBLE))
		affect_from_char(ch, SPELL_INVISIBLE);

	AFF_FLAGS(ch).AFF_INVISIBLE = 0;
	AFF_FLAGS(ch).AFF_HIDE = 0;

	if (GET_LEVEL(ch) < LVL_IMMORT)
		act("$n slowly fades into existence.", FALSE, ch, 0, 0, TO_ROOM);
	else
		act("You feel a strange presence as $n appears, seemingly from nowhere.",
			 FALSE, ch, 0, 0, TO_ROOM);
}


int compute_armor_class(struct char_data *ch)
{
	int armorclass = GET_AC(ch);

	if (AWAKE(ch))
		armorclass += dex_app[GET_DEX(ch)].defensive * 10;

	return (MAX(-100, armorclass));      /* -100 is lowest */
}


void free_messages_type(struct msg_type *msg)
{
	if (msg->attacker_msg)	free(msg->attacker_msg);
	if (msg->victim_msg)		free(msg->victim_msg);
	if (msg->room_msg)		free(msg->room_msg);
}


void free_messages(void)
{
	int i;

	for (i = 0; i < MAX_MESSAGES; i++)
		while (fight_messages[i].msg) {
			struct message_type *former = fight_messages[i].msg;

			free_messages_type(&former->die_msg);
			free_messages_type(&former->miss_msg);
			free_messages_type(&former->hit_msg);
			free_messages_type(&former->god_msg);

			fight_messages[i].msg = fight_messages[i].msg->next;
			free(former);
		}
}


void load_messages(void)
{
	FILE *fl;
	int i, type;
	struct message_type *messages;
	char chk[128];

	if (!(fl = fopen(MESS_FILE, "r"))) {
		log("SYSERR: Error reading combat message file %s: %s", MESS_FILE, strerror(errno));
		exit(1);
	}

	for (i = 0; i < MAX_MESSAGES; i++) {
		fight_messages[i].a_type = 0;
		fight_messages[i].number_of_attacks = 0;
		fight_messages[i].msg = NULL;
	}

	fgets(chk, 128, fl);
	while (!feof(fl) && (*chk == '\n' || *chk == '*'))
		fgets(chk, 128, fl);

	while (*chk == 'M') {
		fgets(chk, 128, fl);
		sscanf(chk, " %d\n", &type);
		for (i = 0; (i < MAX_MESSAGES) && (fight_messages[i].a_type != type) &&
				  (fight_messages[i].a_type); i++);
		if (i >= MAX_MESSAGES) {
			log("SYSERR: Too many combat messages.  Increase MAX_MESSAGES and recompile.");
			exit(1);
		}
		CREATE(messages, struct message_type, 1);
		fight_messages[i].number_of_attacks++;
		fight_messages[i].a_type = type;
		messages->next = fight_messages[i].msg;
		fight_messages[i].msg = messages;

		messages->die_msg.attacker_msg = fread_action(fl, i);
		messages->die_msg.victim_msg = fread_action(fl, i);
		messages->die_msg.room_msg = fread_action(fl, i);
		messages->miss_msg.attacker_msg = fread_action(fl, i);
		messages->miss_msg.victim_msg = fread_action(fl, i);
		messages->miss_msg.room_msg = fread_action(fl, i);
		messages->hit_msg.attacker_msg = fread_action(fl, i);
		messages->hit_msg.victim_msg = fread_action(fl, i);
		messages->hit_msg.room_msg = fread_action(fl, i);
		messages->god_msg.attacker_msg = fread_action(fl, i);
		messages->god_msg.victim_msg = fread_action(fl, i);
		messages->god_msg.room_msg = fread_action(fl, i);
		fgets(chk, 128, fl);
		while (!feof(fl) && (*chk == '\n' || *chk == '*'))
			fgets(chk, 128, fl);
	}

	fclose(fl);
}


void pos_text(struct char_data *ch, struct char_data *victim, int dam)
{
	/* Use send_to_char -- act() doesn't send message if you are DEAD. */

	switch (GET_POS(victim)) {
		case POS_MORTALLYW:
			act("$n is mortally wounded, and will die soon, if not aided.", 
				 TRUE, victim, 0, 0, TO_ROOM);
			send_to_char(victim, "You are mortally wounded, and will die soon, "
							 "if not aided.\r\n");
			break;

		case POS_INCAP:
			act("$n is incapacitated and will slowly die, if not aided.", 
				 TRUE, victim, 0, 0, TO_ROOM);
			send_to_char(victim, "You are incapacitated an will slowly die, "
							 "if not aided.\r\n");
			break;

		case POS_STUNNED:
			act("$n is stunned, but will probably regain consciousness again.", 
				 TRUE, victim, 0, 0, TO_ROOM);
			send_to_char(victim, "You're stunned, but will probably regain "
							 "consciousness again.\r\n");
			break;

		case POS_DEAD:
			act("$n is dead!  R.I.P.", FALSE, victim, 0, 0, TO_ROOM);
			send_to_char(victim, "You are dead!  Sorry...\r\n");
			break;

		default:			/* >= POSITION SLEEPING */
			if (dam > (GET_MAX_HIT(victim) / 4))
				send_to_char(victim, "That really did HURT!\r\n");

			if (GET_HIT(victim) < (GET_MAX_HIT(victim) / 4)) {
				send_to_char(victim, "%sYou wish that your wounds would stop "
								 "BLEEDING so much!%s\r\n",
								 CCRED(victim, C_SPR), CCNRM(victim, C_SPR));
				if (ch != victim && MOB_FLAGGED(victim, MOB_WIMPY))
					do_flee(victim, NULL, 0, 0);
			}
			if (!IS_NPC(victim) && GET_WIMP_LEV(victim) && (victim != ch) &&
				 GET_HIT(victim) < GET_WIMP_LEV(victim) && GET_HIT(victim) > 0) {
				send_to_char(victim, "You wimp out, and attempt to flee!\r\n");
				do_flee(victim, NULL, 0, 0);
			}
			break;
	}
}


void update_pos(struct char_data *victim)
{
	if (GET_HIT(victim) > 0 && GET_POS(victim) > POS_STUNNED)
		return;
	else if (AFF_FLAGGED(victim, AFF_STUN) && GET_HIT(victim) > 0)
		GET_POS(victim) = POS_STUNNED;
	else if (GET_HIT(victim) > 0)
		GET_POS(victim) = POS_STANDING;
	else if (GET_HIT(victim) <= -11)
		GET_POS(victim) = POS_DEAD;
	else if (GET_HIT(victim) <= -6)
		GET_POS(victim) = POS_MORTALLYW;
	else if (GET_HIT(victim) <= -3)
		GET_POS(victim) = POS_INCAP;
	else
		GET_POS(victim) = POS_STUNNED;
}


void check_killer(struct char_data *ch, struct char_data *vict)
{
	if (PLR_FLAGGED(vict, PLR_KILLER) || PLR_FLAGGED(vict, PLR_THIEF))
		return;
	if (PLR_FLAGGED(ch, PLR_KILLER) || IS_NPC(ch) || IS_NPC(vict) || ch == vict)
		return;

	PLR_FLAGS(ch).PLR_KILLER = 1;
	send_to_char(ch, "If you want to be a PLAYER KILLER, so be it...\r\n");
	mudlog(BRF, LVL_IMMORT, TRUE, "PC Killer bit set on %s for initiating "
			 "eattack on %s at %s.", GET_NAME(ch), GET_NAME(vict), 
			 world[IN_ROOM(vict)].name);
}


/* start one char fighting another (yes, it is horrible, I know... )  */
void set_fighting(struct char_data *ch, struct char_data *vict)
{
	if (ch == vict)
		return;

	if (FIGHTING(ch)) {
		core_dump();
		return;
	}

	ch->next_fighting = combat_list;
	combat_list = ch;

	if (AFF_FLAGGED(ch, AFF_SLEEP))
		affect_from_char(ch, SPELL_SLEEP);

	FIGHTING(ch) = vict;
	GET_POS(ch) = POS_FIGHTING;

	if (!pk_allowed)
		check_killer(ch, vict);
}



/* remove a char from the list of fighting chars */
void stop_fighting(struct char_data *ch)
{
	struct char_data *temp;

	if (ch == next_combat_list)
		next_combat_list = ch->next_fighting;

	REMOVE_FROM_LIST(ch, combat_list, next_fighting);
	ch->next_fighting = NULL;
	FIGHTING(ch) = NULL;
	GET_POS(ch) = POS_STANDING;
	update_pos(ch);
}



void make_corpse(struct char_data *ch)
{
	char buf2[MAX_NAME_LENGTH + 64];
	struct obj_data *corpse, *o;
	struct obj_data *copper, *silver, *gold, *platinum;
	int i;

	corpse = create_obj();

	corpse->item_number = NOTHING;
	IN_ROOM(corpse) = NOWHERE;
	corpse->name = strdup("corpse");

	snprintf(buf2, sizeof(buf2), "The corpse of %s is lying here.", GET_NAME(ch));
	corpse->description = strdup(buf2);

	snprintf(buf2, sizeof(buf2), "the corpse of %s", GET_NAME(ch));
	corpse->short_description = strdup(buf2);

	GET_OBJ_TYPE(corpse) = ITEM_CONTAINER;
	GET_OBJ_WEAR(corpse).ITEM_WEAR_TAKE = 1;
	GET_OBJ_EXTRA(corpse).ITEM_NODONATE = 1;
	GET_OBJ_VAL(corpse, 0) = 0;	/* You can't store stuff in a corpse */
	GET_OBJ_VAL(corpse, 3) = 1;	/* corpse identifier */
	GET_OBJ_WEIGHT(corpse) = GET_WEIGHT(ch) + IS_CARRYING_W(ch);
	GET_OBJ_RENT(corpse) = 100000;
	if (IS_NPC(ch))
		GET_OBJ_TIMER(corpse) = max_npc_corpse_time;
	else
		GET_OBJ_TIMER(corpse) = max_pc_corpse_time;

	/* transfer character's inventory to the corpse */
	corpse->contains = ch->carrying;
	for (o = corpse->contains; o != NULL; o = o->next_content)
		IN_OBJ(o) = corpse;
	object_list_new_owner(corpse, NULL);

	/* transfer character's equipment to the corpse */
	for (i = 0; i < NUM_WEARS; i++)
		if (GET_EQ(ch, i))
			obj_to_obj(unequip_char(ch, i), corpse);

	/* transfer gold */
	if (convert_all_to_copper(ch) > 0) {
		/*
		 * following 'if' clause added to fix gold duplication loophole
		 * The above line apparently refers to the old "partially log in,
		 * kill the game character, then finish login sequence" duping
		 * bug. The duplication has been fixed (knock on wood) but the
		 * test below shall live on, for a while. -gg 3/3/2002
		 */
		if (IS_NPC(ch) || ch->desc) {
			if (GET_COPPER(ch)) {
				copper = create_money(0, 0, 0, GET_COPPER(ch), COPPER_COINS);
				obj_to_obj(copper, corpse);
			}

			if (GET_SILVER(ch)) {
				silver = create_money(0, 0, GET_SILVER(ch), 0, SILVER_COINS);
				obj_to_obj(silver, corpse);
			}

			if (GET_GOLD(ch)) {
				gold = create_money(0, GET_GOLD(ch), 0, 0, GOLD_COINS);
				obj_to_obj(gold, corpse);
			}

			if (GET_PLATINUM(ch)) {
				platinum = create_money(GET_PLATINUM(ch), 0, 0, 0, PLAT_COINS);
				obj_to_obj(platinum, corpse);
			}
		}
		GET_PLATINUM(ch) = 0;
		GET_GOLD(ch) = 0;
		GET_SILVER(ch) = 0;
		GET_COPPER(ch) = 0;
	}
	ch->carrying = NULL;
	IS_CARRYING_N(ch) = 0;
	IS_CARRYING_W(ch) = 0;

	obj_to_room(corpse, IN_ROOM(ch));
}


/* When ch kills victim */
void change_alignment(struct char_data *ch, struct char_data *victim)
{
	/*
	 * new alignment change algorithm: if you kill a monster with alignment A,
	 * you move 1/16th of the way to having alignment -A.  Simple and fast.
	 */
	GET_ALIGNMENT(ch) += (-GET_ALIGNMENT(victim) - GET_ALIGNMENT(ch)) / 16;
}



void death_cry(struct char_data *ch)
{
	int door;

	act("Your blood freezes as you hear $n's death cry.", FALSE, ch, 0, 0, TO_ROOM);

	for (door = 0; door < NUM_OF_DIRS; door++)
		if (CAN_GO(ch, door))
			send_to_room(world[IN_ROOM(ch)].dir_option[door]->to_room, "Your blood freezes as you hear someone's death cry.\r\n");
}



void raw_kill(struct char_data *ch, struct char_data *killer)
{
	if (FIGHTING(ch))
		stop_fighting(ch);

	while (ch->affected)
		affect_remove(ch, ch->affected);

	/* death_cry(ch); */
	/* replaced for MOBprog death action */
	if (killer)
      mprog_death_trigger(ch, killer);

	make_fun_body_pieces(ch, killer);
	make_corpse(ch);
	extract_char(ch);
}



void increase_blood(int rm)
{
	RM_BLOOD(rm) = MIN(RM_BLOOD(rm) + 1, 10);
}


void die(struct char_data *ch, struct char_data *killer)
{
	increase_blood(IN_ROOM(ch));

	gain_exp(ch, -(GET_EXP(ch) / 2));
	if (!IS_NPC(ch)) {
		PLR_FLAGS(ch).PLR_KILLER = 0;
		PLR_FLAGS(ch).PLR_THIEF = 0;
	}
	
	raw_kill(ch, killer);
}



void perform_group_gain(struct char_data *ch, int base,
								struct char_data *victim)
{
	int share;

	share = MIN(max_exp_gain, MAX(1, base));

	if (share > 1)
		send_to_char(ch, "You receive your share of experience -- %d points.\r\n", share);
	else
		send_to_char(ch, "You receive your share of experience -- one measly little point!\r\n");

	gain_exp(ch, share);
	change_alignment(ch, victim);
}


void group_gain(struct char_data *ch, struct char_data *victim)
{
	int tot_members, base, tot_gain;
	struct char_data *k;
	struct follow_type *f;

	if (!(k = ch->master))
		k = ch;

	if (AFF_FLAGGED(k, AFF_GROUP) && (IN_ROOM(k) == IN_ROOM(ch)))
		tot_members = 1;
	else
		tot_members = 0;

	for (f = k->followers; f; f = f->next)
		if (AFF_FLAGGED(f->follower, AFF_GROUP) && IN_ROOM(f->follower) == IN_ROOM(ch))
			tot_members++;

	/* round up to the next highest tot_members */
	tot_gain = (GET_EXP(victim) / 3) + tot_members - 1;

	/* prevent illegal xp creation when killing players */
	if (!IS_NPC(victim))
		tot_gain = MIN(max_exp_loss * 2 / 3, tot_gain);

	if (tot_members >= 1)
		base = MAX(1, tot_gain / tot_members);
	else
		base = 0;

	if (AFF_FLAGGED(k, AFF_GROUP) && IN_ROOM(k) == IN_ROOM(ch))
		perform_group_gain(k, base, victim);

	for (f = k->followers; f; f = f->next)
		if (AFF_FLAGGED(f->follower, AFF_GROUP) && IN_ROOM(f->follower) == IN_ROOM(ch))
			perform_group_gain(f->follower, base, victim);
}


void solo_gain(struct char_data *ch, struct char_data *victim)
{
	int exp;

	exp = MIN(max_exp_gain, GET_EXP(victim) / 3);

	/* Calculate level-difference bonus */
	if (IS_NPC(ch))
		exp += MAX(0, (exp * MIN(4, (GET_LEVEL(victim) - GET_LEVEL(ch)))) / 8);
	else
		exp += MAX(0, (exp * MIN(8, (GET_LEVEL(victim) - GET_LEVEL(ch)))) / 8);

	exp = MAX(exp, 1);

	if (exp > 1)
		send_to_char(ch, "You receive %d experience points.\r\n", exp);
	else
		send_to_char(ch, "You receive one lousy experience point.\r\n");

	gain_exp(ch, exp);
	change_alignment(ch, victim);
}


char *replace_string(const char *str, const char *weapon_singular, const char *weapon_plural)
{
	static char buf[256];
	char *cp = buf;

	for (; *str; str++) {
		if (*str == '#') {
			switch (*(++str)) {
				case 'W':
					for (; *weapon_plural; *(cp++) = *(weapon_plural++));
					break;
				case 'w':
					for (; *weapon_singular; *(cp++) = *(weapon_singular++));
					break;
				default:
					*(cp++) = '#';
					break;
			}
		} else
			*(cp++) = *str;

		*cp = 0;
	}				/* For */

	return (buf);
}


/* message for doing damage with a weapon */
void dam_message(int dam, struct char_data *ch, struct char_data *victim,
					  int w_type)
{
	char *buf;
	int msgnum;

	static struct dam_weapon_type {
		const char *to_room;
		const char *to_char;
		const char *to_victim;
	} dam_weapons[] = {

		/* use #w for singular (i.e. "slash") and #W for plural (i.e. "slashes") */

		{
			"$n tries to #w $N, but misses.",	/* 0: 0     */
			"You try to #w $N, but miss.",
			"$n tries to #w you, but misses."
		},

		{
			"$n tickles $N as $e #W $M.",	/* 1: 1..2  */
			"You tickle $N as you #w $M.",
			"$n tickles you as $e #W you."
		},

		{
			"$n barely #W $N.",		/* 2: 3..4  */
			"You barely #w $N.",
			"$n barely #W you."
		},

		{
			"$n #W $N.",			/* 3: 5..6  */
			"You #w $N.",
			"$n #W you."
		},

		{
			"$n #W $N hard.",			/* 4: 7..10  */
			"You #w $N hard.",
			"$n #W you hard."
		},

		{
			"$n #W $N very hard.",		/* 5: 11..14  */
			"You #w $N very hard.",
			"$n #W you very hard."
		},

		{
			"$n #W $N extremely hard.",	/* 6: 15..19  */
			"You #w $N extremely hard.",
			"$n #W you extremely hard."
		},

		{
			"$n massacres $N to small fragments with $s #w.",	/* 7: 19..23 */
			"You massacre $N to small fragments with your #w.",
			"$n massacres you to small fragments with $s #w."
		},

		{
			"$n OBLITERATES $N with $s deadly #w!!",	/* 8: > 23   */
			"You OBLITERATE $N with your deadly #w!!",
			"$n OBLITERATES you with $s deadly #w!!"
		}
	};


	w_type -= TYPE_HIT;		/* Change to base of table with text */

	if (dam == 0)		msgnum = 0;
	else if (dam <= 2)    msgnum = 1;
	else if (dam <= 4)    msgnum = 2;
	else if (dam <= 6)    msgnum = 3;
	else if (dam <= 10)   msgnum = 4;
	else if (dam <= 14)   msgnum = 5;
	else if (dam <= 19)   msgnum = 6;
	else if (dam <= 23)   msgnum = 7;
	else			msgnum = 8;

	/* damage message to onlookers */
	buf = replace_string(dam_weapons[msgnum].to_room,
								attack_hit_text[w_type].singular, attack_hit_text[w_type].plural);
	act(buf, FALSE, ch, NULL, victim, TO_NOTVICT);

	/* damage message to damager */
	send_to_char(ch, CCYEL(ch, C_CMP));
	buf = replace_string(dam_weapons[msgnum].to_char,
								attack_hit_text[w_type].singular, attack_hit_text[w_type].plural);
	act(buf, FALSE, ch, NULL, victim, TO_CHAR);
	send_to_char(ch, CCNRM(ch, C_CMP));

	/* damage message to damagee */
	send_to_char(victim, CCRED(victim, C_CMP));
	buf = replace_string(dam_weapons[msgnum].to_victim,
								attack_hit_text[w_type].singular, attack_hit_text[w_type].plural);
	act(buf, FALSE, ch, NULL, victim, TO_VICT | TO_SLEEP);
	send_to_char(victim, CCNRM(victim, C_CMP));
}


/*
 *  message for doing damage with a spell or skill
 *  C3.0: Also used for weapon damage on miss and death blows
 */
int skill_message(int dam, struct char_data *ch, struct char_data *vict,
						int attacktype)
{
	int i, j, nr;
	struct message_type *msg;

	struct obj_data *weap = GET_EQ(ch, WEAR_WIELD);

	for (i = 0; i < MAX_MESSAGES; i++) {
		if (fight_messages[i].a_type == attacktype) {
			nr = dice(1, fight_messages[i].number_of_attacks);
			for (j = 1, msg = fight_messages[i].msg; (j < nr) && msg; j++)
				msg = msg->next;

			if (!IS_NPC(vict) && (GET_LEVEL(vict) >= LVL_IMMORT)) {
				act(msg->god_msg.attacker_msg, FALSE, ch, weap, vict, TO_CHAR);
				act(msg->god_msg.victim_msg, FALSE, ch, weap, vict, TO_VICT);
				act(msg->god_msg.room_msg, FALSE, ch, weap, vict, TO_NOTVICT);
			} else if (dam != 0) {
				/*
				 * Don't send redundant color codes for TYPE_SUFFERING & other types
				 * of damage without attacker_msg.
				 */
				if (GET_POS(vict) == POS_DEAD) {
					if (msg->die_msg.attacker_msg) {
						send_to_char(ch, CCYEL(ch, C_CMP));
						act(msg->die_msg.attacker_msg, FALSE, ch, weap, vict, TO_CHAR);
						send_to_char(ch, CCNRM(ch, C_CMP));
					}

					send_to_char(vict, CCRED(vict, C_CMP));
					act(msg->die_msg.victim_msg, FALSE, ch, weap, vict, TO_VICT | TO_SLEEP);
					send_to_char(vict, CCNRM(vict, C_CMP));

					act(msg->die_msg.room_msg, FALSE, ch, weap, vict, TO_NOTVICT);
				} else {
					if (msg->hit_msg.attacker_msg) {
						send_to_char(ch, CCYEL(ch, C_CMP));
						act(msg->hit_msg.attacker_msg, FALSE, ch, weap, vict, TO_CHAR);
						send_to_char(ch, CCNRM(ch, C_CMP));
					}

					send_to_char(vict, CCRED(vict, C_CMP));
					act(msg->hit_msg.victim_msg, FALSE, ch, weap, vict, TO_VICT | TO_SLEEP);
					send_to_char(vict, CCNRM(vict, C_CMP));

					act(msg->hit_msg.room_msg, FALSE, ch, weap, vict, TO_NOTVICT);
				}
			} else if (ch != vict) {	/* Dam == 0 */
				if (msg->miss_msg.attacker_msg) {
					send_to_char(ch, CCYEL(ch, C_CMP));
					act(msg->miss_msg.attacker_msg, FALSE, ch, weap, vict, TO_CHAR);
					send_to_char(ch, CCNRM(ch, C_CMP));
				}

				send_to_char(vict, CCRED(vict, C_CMP));
				act(msg->miss_msg.victim_msg, FALSE, ch, weap, vict, TO_VICT | TO_SLEEP);
				send_to_char(vict, CCNRM(vict, C_CMP));

				act(msg->miss_msg.room_msg, FALSE, ch, weap, vict, TO_NOTVICT);
			}
			return (1);
		}
	}
	return (0);
}


/*
 * Alert: As of bpl14, this function returns the following codes:
 *	< 0	Victim died.
 *	= 0	No damage.
 *	> 0	How much damage done.
 */
int damage(struct char_data *ch, struct char_data *victim, int dam, int attacktype)
{
	int goal, roll;
	long local_gold = 0;
	char local_buf[256];
	int missile = FALSE;
	
	if (GET_POS(victim) <= POS_DEAD) {
		/* This is "normal"-ish now with delayed extraction. -gg 3/15/2001 */
		if (PLR_FLAGGED(victim, PLR_NOTDEADYET) || MOB_FLAGGED(victim, MOB_NOTDEADYET))
			return (-1);

		log("SYSERR: Attempt to damage corpse '%s' in room #%d by '%s'.",
			 GET_NAME(victim), GET_ROOM_VNUM(IN_ROOM(victim)), GET_NAME(ch));
		die(victim, ch);
		return (-1);			/* -je, 7/7/92 */
	}

	if (IN_ROOM(ch) != IN_ROOM(victim))
		missile = TRUE;

	/* peaceful rooms */
	if (ch != victim && ROOM_FLAGGED(IN_ROOM(ch), ROOM_PEACEFUL)) {
		send_to_char(ch, "This room just has such a peaceful, easy feeling...\r\n");
		return (0);
	}

	if (pk_allowed && PRF_FLAGGED(victim, PRF_AFK)) {
		check_killer(ch, victim);
		send_to_char(ch, "No attacking people who are AFK!\r\n");
		return (0);
	}

	/* shopkeeper protection */
	if (!ok_damage_shopkeeper(ch, victim))
		return (0);

	/* You can't damage an immortal! */
	if (!IS_NPC(victim) && (GET_LEVEL(victim) >= LVL_IMMORT))
		dam = 0;

	if (victim != ch && !missile) {
		/* Start the attacker fighting the victim */
		if (GET_POS(ch) > POS_STUNNED && (FIGHTING(ch) == NULL))
			set_fighting(ch, victim);

		/* Start the victim fighting the attacker */
		if (GET_POS(victim) > POS_STUNNED && (FIGHTING(victim) == NULL)) {
			set_fighting(victim, ch);
			if (MOB_FLAGGED(victim, MOB_MEMORY) && !IS_NPC(ch))
				remember(victim, ch);
		}
	}

	/* If you attack a pet, it hates your guts */
	if (victim->master == ch)
		stop_follower(victim);

	/* If the attacker is invisible, he becomes visible */
	if (AFF_FLAGGED(ch, AFF_INVISIBLE) || AFF_FLAGGED(ch, AFF_HIDE))
		appear(ch);

	if (!IS_NPC(ch) && GET_SKILL(ch, SKILL_ENHANCED_DMG)) {
		/* Actually, if we just did a straight check on the skill,
		 * warriors would always get double damage.  Better to cut
		 * it down some.  -ejg  */
		if (rand_number(1, 101) > (GET_SKILL(ch, SKILL_ENHANCED_DMG) / 5))
			dam *= 2;
	}

	if (PLR_FLAGGED(victim, PLR_FISHING))
		dam *= 1.5;

	/* Cut damage in half if victim has sanct, to a minimum 1 */
	if (AFF_FLAGGED(victim, AFF_SANCTUARY) && dam >= 2)
		dam /= 2;

	/* Check if victim is mirror imaged */
	if (AFF_FLAGGED(victim, AFF_MIRROR_IMAGE)) {
		if (dice(1, (GET_LEVEL(victim)/5)) != 1) {
			act("You miss $N, and hit a mirrored image.", 
				 TRUE, ch, 0, victim, TO_CHAR);

			act("$n's attack misses you, and hits a mirror image.", 
				 TRUE, ch, 0, victim, TO_VICT);

			return(0);
		}
	}

	/* Check for PK if this is not a PK MUD */
	if (!pk_allowed) {
		check_killer(ch, victim);
		if (PLR_FLAGGED(ch, PLR_KILLER) && (ch != victim))
			dam = 0;
	}

	/* Set the maximum damage per round and subtract the hit points */
	dam = MAX(MIN(dam, 100), 0);
	GET_HIT(victim) -= dam;

	/* Gain exp for the hit */
	if (ch != victim && !missile)
		gain_exp(ch, GET_LEVEL(victim) * dam);

	update_pos(victim);

	/*
	 * skill_message sends a message from the messages file in lib/misc.
	 * dam_message just sends a generic "You hit $n extremely hard.".
	 * skill_message is preferable to dam_message because it is more
	 * descriptive.
	 * 
	 * If we are _not_ attacking with a weapon (i.e. a spell), always use
	 * skill_message. If we are attacking with a weapon: If this is a miss or a
	 * death blow, send a skill_message if one exists; if not, default to a
	 * dam_message. Otherwise, always send a dam_message.
	 */
   if (attacktype != -1) {
		if (!IS_WEAPON(attacktype) || missile)
			skill_message(dam, ch, victim, attacktype);
		else {
			if (GET_POS(victim) == POS_DEAD || dam == 0) {
				if (!skill_message(dam, ch, victim, attacktype))
					dam_message(dam, ch, victim, attacktype);
			} else {
				dam_message(dam, ch, victim, attacktype);
			}
		}
	}
	
	pos_text(ch, victim, dam);

	/* Help out poor linkless people who are attacked */
	if (!IS_NPC(victim) && !(victim->desc) && GET_POS(victim) > POS_STUNNED) {
		do_flee(victim, NULL, 0, 0);
		if (!FIGHTING(victim)) {
			act("$n is rescued by divine forces.", FALSE, victim, 0, 0, TO_ROOM);
			GET_WAS_IN(victim) = IN_ROOM(victim);
			char_from_room(victim);
			char_to_room(victim, 0);
		}
	}

	/* stop someone from fighting if they're stunned or worse */
	if (GET_POS(victim) <= POS_STUNNED && FIGHTING(victim) != NULL)
		stop_fighting(victim);

	if (!IS_NPC(victim))
		goal = GET_SKILL(victim, SKILL_LUCK);
	else
		goal = 0;

	roll = rand_number(1, 101);

	if (GET_POS(victim) < POS_STUNNED && GET_POS(victim) > POS_DEAD && 
		 FIGHTING(victim) && IS_NPC(ch) && roll <= goal) {
		act("$N stops beating you up and walks away.  It must be your lucky "
			 "day!", FALSE, victim, 0, ch, TO_CHAR);
		act("$N stops beating up $n and walks away.  It must be $s lucky day!",
			 FALSE, victim, 0, ch, TO_NOTVICT);
		stop_fighting(victim);
		stop_fighting(ch);
	}
	
	/* Uh oh.  Victim died. */
	if (GET_POS(victim) == POS_DEAD) {
		if (ch != victim && (IS_NPC(victim) || victim->desc)) {
			if (!missile) {
				if (AFF_FLAGGED(ch, AFF_GROUP))
					group_gain(ch, victim);
				else
					solo_gain(ch, victim);
			}
		}

		if (!IS_NPC(victim)) {
			mudlog(BRF, LVL_IMMORT, TRUE, "%s killed by %s at %s", 
					 GET_NAME(victim), GET_NAME(ch), world[IN_ROOM(victim)].name);

			if (IS_NPC(ch))
				brag(ch, victim);
			if (MOB_FLAGGED(ch, MOB_MEMORY))
				forget(ch, victim);
		}
		/* Cant determine GET_GOLD on corpse, so do now and store */
		if (IS_NPC(victim)) {
			local_gold = GET_GOLD(victim);
			snprintf(local_buf, sizeof(local_buf), "%ld", (long)local_gold);
		}

		die(victim, ch);

		if (!missile) { /* Can't loot or stuff if the corpse isn't here! */

			/* If you don't want aggressive mobs to loot, or mobs to loot
			 * other mobs, just get rid of the following lines.  It *is* a
			 * pretty evil thing to do, but then flagging a mob aggressive
			 * isn't nice either...   -ejg */
			if (IS_NPC(ch) &&
				 (MOB_FLAGGED(ch, MOB_AGGRESSIVE) ||
				  MOB_FLAGGED(ch, MOB_AGGR_EVIL) ||
				  MOB_FLAGGED(ch, MOB_AGGR_NEUTRAL) ||
				  MOB_FLAGGED(ch, MOB_AGGR_GOOD)) ||
				 IS_NPC(victim))
				do_get(ch, "all corpse", 0, 0);
		
			/* If Autoloot enabled, get all corpse */
			else if (!IS_NPC(ch) && IS_NPC(victim) && PRF_FLAGGED(ch, PRF_AUTOLOOT))
				do_get(ch, "all corpse", 0, 0);

			/* If Autoloot AND AutoSplit AND we got money, split with group */
			if (AFF_FLAGGED(ch, AFF_GROUP) && local_gold > 0 &&
				 PRF_FLAGGED(ch, PRF_AUTOSPLIT) && PRF_FLAGGED(ch, PRF_AUTOLOOT)) {
				do_split(ch, local_buf, 0, 0);
			}

			/* If Autosac enabled, get all corpse */
			if (IS_NPC(victim) && !IS_NPC(ch) && PRF_FLAGGED(ch, PRF_AUTOSAC)) {
				if (IS_EVIL(ch))
					do_sac(ch, "corpse", 0, SCMD_SACRIFICE);
				else if (IS_GOOD(ch))
					do_sac(ch, "corpse", 0, SCMD_LASTRITES);
				else
					do_sac(ch, "corpse", 0, SCMD_BURIAL);
			}
		}
		
		return (-1);
	}
	return (dam);
}


int racial_combat_modifiers(struct char_data *ch, struct char_data *victim)
{
	int calc_thaco = 0;
	
	/* Elves and halflings get bonuses with certain weapons */
	if (IS_ELF(ch))  {
		struct obj_data *wielded = GET_EQ(ch, WEAR_WIELD);
		if (wielded) {
			if (isname("sword", wielded->name) || isname("bow", wielded->name))
				calc_thaco += 1;
		}
	}

	return (calc_thaco);
}

/*
 * Calculate the THAC0 of the attacker.
 */
int compute_thaco(struct char_data *ch, struct char_data *victim)
{
	int calc_thaco;

	if (!IS_NPC(ch))
		calc_thaco = thaco(GET_CLASS(ch), GET_LEVEL(ch));
	else		/* THAC0 for monsters is set in the HitRoll */
		calc_thaco = 20;

	calc_thaco -= str_app[STRENGTH_APPLY_INDEX(ch)].tohit;
	calc_thaco -= GET_HITROLL(ch);
	calc_thaco -= (int) ((GET_INT(ch) - 13) / 1.5);	/* Intelligence helps! */
	calc_thaco -= (int) ((GET_WIS(ch) - 13) / 1.5);	/* So does wisdom */

	if (IS_EVIL(ch) && AFF_FLAGGED(victim, AFF_PROTECT_EVIL))
		calc_thaco += 2; /* harder to hit :p   -ejg */
	else if (IS_GOOD(ch) && AFF_FLAGGED(victim, AFF_PROTECT_GOOD))
		calc_thaco += 2;

	calc_thaco -= racial_combat_modifiers(ch, victim);

	/* Harder to fight in bad weather, if you are outdoors */
	if (OUTSIDE(ch) &&
		 (zone_table[IN_ZONE(ch)].sky == SKY_RAINING ||
		  zone_table[IN_ZONE(ch)].sky == SKY_LIGHTNING))
		calc_thaco += 1;

	return calc_thaco;
}


void weapon_spells(struct char_data *ch, struct char_data *vict, struct obj_data *wpn)
{
	int i = 0, rndm;

	if (wpn && HAS_SPELLS(wpn)) {
		while (GET_WEAPON_SPELL(wpn, i) && i < MAX_WEAPON_SPELLS) {
			/* Do some sanity checking, in case someone flees, etc. */
			/* Can't have enough of 'em.  */

			if (IN_ROOM(ch) != IN_ROOM(vict)) {
				if (FIGHTING(ch) && FIGHTING(ch) == vict)
					stop_fighting(ch);
				return;
			}

			rndm = rand_number(1, 100);

			if (GET_WEAPON_SPELL_PCT(wpn, i) >= rndm) {
				act("$p leaps to action with an attack of its own.",
					 TRUE, ch, wpn, 0, TO_CHAR);
				act("$p leaps to action with an attack of its own.",
					 TRUE, ch, wpn, 0, TO_ROOM);

				if (call_magic(ch, vict, NULL, GET_WEAPON_SPELL(wpn, i),
									GET_WEAPON_SPELL_LVL(wpn, i), CAST_WAND, 0) < 0)
					return; /* victim died */
			}
			i++;
		}
	}
}


int hit_sanity_check(struct char_data *ch, struct char_data *victim)
{
	/* Do some sanity checking, in case someone flees, etc. */
	if (IN_ROOM(ch) != IN_ROOM(victim)) {
		if (FIGHTING(ch) && FIGHTING(ch) == victim)
			stop_fighting(ch);
		return (1);
	}
	return (0);
}


int find_weapon_type(struct char_data *ch, int dual)
{
	struct obj_data *wielded;
	int w_type;
	
	if (dual) 
		wielded = GET_EQ(ch, WEAR_DWIELD);
	else
		wielded = GET_EQ(ch, WEAR_WIELD);

	/* Find the weapon type (for display purposes only) */
	if (wielded && GET_OBJ_TYPE(wielded) == ITEM_WEAPON)
		w_type = GET_OBJ_VAL(wielded, 3) + TYPE_HIT;
	else {
		if (IS_NPC(ch) && ch->mob_specials.attack_type != 0)
			w_type = ch->mob_specials.attack_type + TYPE_HIT;
		else
			w_type = TYPE_HIT;
	}

	return w_type;
}


int determine_hit(struct char_data *ch, struct char_data *victim)
{
	int dam, calc_thaco, victim_ac, diceroll;
	
	/* Calculate chance of hit. Lower THAC0 is better for attacker. */
	calc_thaco = compute_thaco(ch, victim);

	/* Dual wielding incurs penalties... */
	if (GET_EQ(ch, WEAR_DWIELD))
		calc_thaco += 2;

	/* Calculate the raw armor including magic armor.  Lower AC is
	 * better for defender. */
	victim_ac = compute_armor_class(victim) / 10;

	/* roll the die and take your chances... */
	diceroll = rand_number(1, 20);

	/*
	 * Decide whether this is a hit or a miss.
	 *
	 *  Victim asleep = hit, otherwise:
	 *     1   = Automatic miss.
	 *   2..19 = Checked vs. AC.
	 *    20   = Automatic hit.
	 */
	if (diceroll == 20 || !AWAKE(victim))
		dam = TRUE;
	else if (diceroll == 1)
		dam = FALSE;
	else
		dam = (calc_thaco - diceroll <= victim_ac);

	return dam;
}


void swing_and_a_miss(struct char_data *ch, struct char_data *victim, 
							 int type, int w_type)
{
	/* the attacker missed the victim */
	damage(ch, victim, 0, type == SKILL_BACKSTAB ? SKILL_BACKSTAB : w_type);
}


int calc_damage(struct char_data *ch, struct char_data *victim, 
					 struct obj_data *wielded)
{
	/* okay, we know the guy has been hit.  now calculate damage. */

	/* Start with the damage bonuses: the damroll and strength apply */
	int dam = str_app[STRENGTH_APPLY_INDEX(ch)].todam;
	dam += GET_DAMROLL(ch);

	/* Maybe holding arrow? */
	if (wielded && GET_OBJ_TYPE(wielded) == ITEM_WEAPON) {
		/* Add weapon-based damage if a weapon is being wielded */
		dam += dice(GET_OBJ_VAL(wielded, 1), GET_OBJ_VAL(wielded, 2));
	} else {
		/* If no weapon, add bare hand damage instead */
		if (IS_NPC(ch))
			dam += dice(ch->mob_specials.damnodice, ch->mob_specials.damsizedice);
		else {
			if (rand_number(1, 101) > GET_SKILL(ch, SKILL_UNARMED_COMBAT))
				dam += rand_number(0, 2);	/* Max 2 bare hand damage for players */
			else
				dam += rand_number(1, GET_LEVEL(ch));
		}
	}

	/*
	 * Include a damage multiplier if victim isn't ready to fight:
	 *
	 * Position sitting  1.33 x normal
	 * Position resting  1.66 x normal
	 * Position sleeping 2.00 x normal
	 * Position stunned  2.33 x normal
	 * Position incap    2.66 x normal
	 * Position mortally 3.00 x normal
	 *
	 * Note, this is a hack because it depends on the particular
	 * values of the POSITION_XXX constants.
	 */
	if (GET_POS(victim) < POS_FIGHTING)
		dam *= 1 + (POS_FIGHTING - GET_POS(victim)) / 3;

	/* at least 1 hp damage min per hit */
	dam = MAX(1, dam);

	return dam;
}




void hit(struct char_data *ch, struct char_data *victim, int type)
{
	struct obj_data *wielded = GET_EQ(ch, WEAR_WIELD);
	int w_type, dam, diceroll, did_hit = FALSE;

	if (hit_sanity_check(ch, victim))
		return;
	
	w_type = find_weapon_type(ch, FALSE);

	if (0 == (dam = determine_hit(ch, victim)))
		 swing_and_a_miss(ch, victim, type, w_type);
	else {
		dam = calc_damage(ch, victim, wielded);
			
		if (type == SKILL_BACKSTAB)
			damage(ch, victim, dam * backstab_mult(GET_LEVEL(ch)), SKILL_BACKSTAB);
		else {
			damage(ch, victim, dam, w_type);
			did_hit = TRUE;

			/* Dual wielders gets another swing! */
			if (GET_EQ(ch, WEAR_DWIELD) && GET_HIT(victim) > 0)
			{
				/* Only double strike with a dual wield weapon upon a
				 * successful skill check. */
				diceroll = dice(1, 100);
				if (diceroll <= GET_SKILL(ch, SKILL_DUAL_WIELD))
					double_hit(ch, FIGHTING(ch), TYPE_UNDEFINED);
			}
		}
	}

	if (did_hit && (type != SKILL_BACKSTAB)) 
		weapon_spells(ch, victim, wielded);
}


void double_hit(struct char_data * ch, struct char_data * victim, int type)
{
	struct obj_data *wielded = GET_EQ(ch, WEAR_DWIELD);
	int w_type, dam, did_hit = FALSE;

	if (hit_sanity_check(ch, victim))
		return;

	w_type = find_weapon_type(ch, TRUE);

	if (0 == (dam = determine_hit(ch, victim)))
		 swing_and_a_miss(ch, victim, type, w_type);
	else {
		dam = calc_damage(ch, victim, wielded);
		did_hit = TRUE;

		if (type == SKILL_BACKSTAB)
			damage(ch, victim, dam * backstab_mult(GET_LEVEL(ch)), SKILL_BACKSTAB);
		else 
			damage(ch, victim, dam, w_type);
	}

	if (did_hit && (type != SKILL_BACKSTAB)) 
		weapon_spells(ch, victim, wielded);
}


int weapon_special(struct obj_data *wpn, struct char_data *ch)
{
   int (*name)(struct char_data *ch, void *me, int cmd, char *argument);
   
   name = obj_index[GET_OBJ_RNUM(wpn)].func;

   if (name != blind_weapon &&
		 name != fireball_weapon &&
       name != curse_weapon)
		return 0;

   return (name)(ch, wpn, 0, "");
}
  

/* control the fights going on.  Called every 2 seconds from comm.c. */
void perform_violence(void)
{
	struct char_data *ch, *vict, *tch;
   struct obj_data *wpn;
   int apr;

	for (ch = combat_list; ch; ch = next_combat_list) {
		next_combat_list = ch->next_fighting;
		apr = 0;

		if (FIGHTING(ch) == NULL || IN_ROOM(ch) != IN_ROOM(FIGHTING(ch))) {
			stop_fighting(ch);
			continue;
		}

		if (IS_NPC(ch)) {
			if (GET_MOB_WAIT(ch) > 0) {
				GET_MOB_WAIT(ch) -= PULSE_VIOLENCE;
				continue;
			}
			GET_MOB_WAIT(ch) = 0;
			if (GET_POS(ch) < POS_FIGHTING) {
				GET_POS(ch) = POS_FIGHTING;
				act("$n scrambles to $s feet!", TRUE, ch, 0, 0, TO_ROOM);
			}
		}

		if ((GET_POS (FIGHTING(ch)) == POS_MORTALLYW) &&
			 (MOB_FLAGGED(ch, MOB_MERCY))) {
			stop_fighting(FIGHTING(ch));
			stop_fighting(ch);
		}

		if (GET_POS(ch) < POS_FIGHTING) {
			send_to_char(ch, "You can't fight while sitting!!\r\n");
			continue;
		}

		if (!(wpn = GET_EQ(ch, WEAR_WIELD)) || 
			 GET_OBJ_TYPE(wpn) != ITEM_WEAPON ||
			 !weapon_special(wpn, ch)) {

			if (!IS_NPC(ch))
			{
				if (GET_SKILL(ch, SKILL_SECOND_ATTACK) >= rand_number(1, 101)) {
					if (GET_SKILL(ch, SKILL_THIRD_ATTACK) >= rand_number(1, 201))
						apr++;

					apr++;
				}
			} else
				apr = ch->mob_specials.num_attacks - 1;
			   /* -1 because an apr of zero means 1 attack. */
			
			/* increment apr by one for every attack they are supposed to
			// get, for the multiple attack skills, you should make sure
			// they only get a subsequent attack if they properly got the
			// previous one.  For instance, you only get third attack if
			// you are getting a second attack.  This doesn't need to be
			// skill based, you can easily make it based upon
			// class/level... see the second example below.
			//
			//   if (AFF_FLAGGED(ch, AFF_HASTE))
			//     apr += number(0, 2);
			//    
			//   if (GET_CLASS(ch) == CLASS_WARRIOR && GET_LEVEL(ch) >= 10)
			//     apr++;
			//
			// If apr is negative they get no attacks, if apr is 0 they
			// get one attack.  APR has a range of -1 to 4, giving a
			// minimum of no attacks, to a maximum of 4.  See the below
			// line for changing that (eg., MAX(-1, MIN(apr, 6)) means a
			// max of 6). */
     
			apr = MAX(-1, MIN(apr, 4));
  
			if (apr >= 0) {
				for (; apr >= 0 && FIGHTING(ch); apr--)
					hit(ch, FIGHTING(ch), TYPE_UNDEFINED);

				if (MOB_FLAGGED(ch, MOB_SPEC) && GET_MOB_SPEC(ch) && 
					 !MOB_FLAGGED(ch, MOB_NOTDEADYET)) {
					char actbuf[MAX_INPUT_LENGTH] = "";
					(GET_MOB_SPEC(ch)) (ch, ch, 0, actbuf);
				}
			}
		}
		
		if (MOB_FLAGGED(ch, MOB_SMART)) {  /* LJ IQ fight */
			vict = NULL;
			for (tch = world[IN_ROOM(ch)].people; tch; tch = tch->next_in_room) {
				if (FIGHTING(tch) == ch && !IS_NPC(tch))
					if (vict == NULL || GET_HIT(tch) < GET_HIT(vict))
						vict = tch;
			}
			
			if (vict && FIGHTING(ch) != vict) {
				stop_fighting(ch);
				hit(ch, vict, TYPE_UNDEFINED);
			}
		}
		
		mprog_hitprcnt_trigger(ch, FIGHTING(ch));
		mprog_fight_trigger(ch, FIGHTING(ch));
	} /* for (ch = combat_list; ch; ch = next_combat_list) { */
}


void strike_missile(struct char_data *ch, struct char_data *tch, 
                    struct obj_data *missile, int dir, int attacktype)
{
	char buf[MAX_STRING_LENGTH];
   int dam = dice(GET_OBJ_VAL(missile, 1), GET_OBJ_VAL(missile, 2));
   dam += GET_DAMROLL(ch);

	/* Not all missile objs give the str damage bonus... */
   if (GET_OBJ_TYPE(missile) == ITEM_THROW ||
		 GET_OBJ_TYPE(missile) == ITEM_ARROW ||
		 GET_OBJ_TYPE(missile) == ITEM_ROCK)
		dam += str_app[STRENGTH_APPLY_INDEX(ch)].todam;
              
   damage(ch, tch, dam, attacktype);
              
   send_to_char(ch, "You hit!\r\n");
   sprintf(buf, "$P flies in from the %s and strikes %s.", 
           dirs[rev_dir[dir]], GET_NAME(tch));
   act(buf, FALSE, tch, 0, missile, TO_ROOM);

   sprintf(buf, "$P flies in from the %s and hits YOU!",
			  dirs[rev_dir[dir]]);
   act(buf, FALSE, tch, 0, missile, TO_CHAR);
} 


void miss_missile(struct char_data *ch, struct char_data *tch, 
						struct obj_data *missile, int dir, int attacktype)
{
	char buf[MAX_STRING_LENGTH];
	
   sprintf(buf, "$P flies in from the %s and hits the ground!", dirs[rev_dir[dir]]);
   act(buf, FALSE, tch, 0, missile, TO_ROOM);
   act(buf, FALSE, tch, 0, missile, TO_CHAR);
   send_to_char(ch, "You missed!\r\n");
}
 

void mob_reaction(struct char_data *ch, struct char_data *vict, int dir)
{
   if (IS_NPC(vict) && !FIGHTING(vict) && GET_POS(vict) > POS_STUNNED) {
      /* can remember so charge! */
		if (MOB_FLAGGED(vict, MOB_MEMORY)) {
			remember(vict, ch);
			act("$n bellows in pain!", FALSE, vict, 0, 0, TO_ROOM); 
			if (GET_POS(vict) == POS_STANDING) {
				if (!do_simple_move(vict, rev_dir[dir], 1))
					act("$n stumbles while trying to run!",
						 FALSE, vict, 0, 0, TO_ROOM);
			} else
				GET_POS(vict) = POS_STANDING;
       
			/* can't remember so try to run away */
		} else
			do_flee(vict, "", 0, 0);
	}
}


void fire_missile(struct char_data *ch, char arg1[MAX_INPUT_LENGTH],
						struct obj_data *missile, int pos, int range, int dir)
{
   int shot = FALSE, found = FALSE, dam = 0;
   int attacktype, room = IN_ROOM(ch), nextroom, distance;
	char buf[MAX_STRING_LENGTH];
	int type = GET_OBJ_TYPE(missile);
   struct char_data *vict;
     
   if (ROOM_FLAGGED(room, ROOM_PEACEFUL)) {
		send_to_char(ch, "This room just has such a peaceful, easy "
						 "feeling...\r\n");
		return;
   }
 
   if (CAN_GO2(room, dir))
		nextroom = EXIT2(room, dir)->to_room;
   else
		nextroom = NOWHERE;
   
   if (type == ITEM_GRENADE) {
		send_to_char(ch, "You throw it!\r\n");
		sprintf(buf, "$n throws %s %s.", 
				  GET_EQ(ch, WEAR_WIELD)->short_description, dirs[dir]);
		act(buf, FALSE, ch, 0, 0, TO_ROOM);

		send_to_room(nextroom, "%s flies in from the %s.\r\n",
						 missile->short_description, dirs[rev_dir[dir]]);

		obj_to_room(unequip_char(ch, pos), nextroom);
		return;
   }

   for (distance = 1; (nextroom != NOWHERE && distance <= range); distance++) {
		for (vict = world[nextroom].people; vict ; vict= vict->next_in_room) {
			if ((isname(arg1, GET_NAME(vict))) && (CAN_SEE(ch, vict))) {
				found = TRUE;
				break;
			}
		}
 
		if (found == 1) {
			/* Daniel Houghton's missile modification */
			if (missile && ROOM_FLAGGED(IN_ROOM(vict), ROOM_PEACEFUL)) {
				send_to_char(ch, "Nah.  Leave them in peace.\r\n");
				return;
			}

			attacktype = GET_OBJ_VAL(missile, 1) + TYPE_HIT;
 
			switch (type) {
				case ITEM_THROW:
					send_to_char(ch, "You throw it!\r\n");
					sprintf(buf, "$n throws %s %s.",
							  GET_EQ(ch,WEAR_WIELD)->short_description, dirs[dir]);
					break;
				case ITEM_ARROW:
					send_to_char(ch, "You aim and fire!\r\n");
					act("$n aims and fires!", TRUE, ch, 0, 0, TO_ROOM);
					break;
				case ITEM_ROCK:
					send_to_char(ch, "You aim and fire!\r\n");
					act("$n aims and fires!", TRUE, ch, 0, 0, TO_ROOM);
					break;
				case ITEM_BOLT:
					send_to_char(ch, "You aim and fire!\r\n");
					act("$n aims and fires!", TRUE, ch, 0, 0, TO_ROOM);
					break;
				default:
					attacktype = TYPE_UNDEFINED;
					break;
			}
 
			if (attacktype == TYPE_UNDEFINED) { /* Huh? */
				log("SYSERR: attacktype undefined in fire_missile().");
				shot = FALSE;
			}
			
			if (0 == (dam = determine_hit(ch, vict)))
				shot = FALSE; /* missed! */
			else
				shot = TRUE;
 
			if (shot == TRUE) {
				strike_missile(ch, vict, missile, dir, attacktype);

				if (rand_number(0, 1) || attacktype == SKILL_THROW)
					obj_to_char(unequip_char(ch, pos), vict);
				else
					extract_obj(unequip_char(ch, pos));
			} else {
				/* ok missed so move missile into new room */
				miss_missile(ch, vict, missile, dir, attacktype);

				if (!rand_number(0, 2) || attacktype == ITEM_THROW)
					obj_to_room(unequip_char(ch, pos), IN_ROOM(vict));
				else
					extract_obj(unequip_char(ch, pos));
			}
 
			/* either way mob remembers */
			mob_reaction(ch, vict, dir);
			WAIT_STATE(ch, PULSE_VIOLENCE);
			return; 
		} 
 
		room = nextroom;
		if (CAN_GO2(room, dir))
			nextroom = EXIT2(room, dir)->to_room;
		else
			nextroom = NOWHERE;
   }
   send_to_char(ch, "Can't find your target!\r\n");
}


void tick_grenade(void)
{
   struct obj_data *i, *tobj, *next;
   struct char_data *tch, *next_tch;
   int s, dam, door;
	room_rnum t;
	
   /* Grenades are activated by pulling the pin - ie, setting the one
	 * of the extra flag bits. After the pin is pulled the grenade
	 * starts counting down. once it reaches zero, it explodes. */
 
   for (i = object_list; i; i = next) { 
		next = i->next;
		if (GET_OBJ_EXTRA(i).ITEM_LIVE_GRENADE == 1) {
			/* update ticks */
			if (GET_OBJ_VAL(i, 0) > 0)
				GET_OBJ_VAL(i, 0) -= 1;
			else { 
				t = NOWHERE;
 
				/* Blow it up! **KABOOM!** */

				/* Checks to see if inside containers.  To avoid possible
				 * infinite loop add a counter variable. We'll jump out
				 * after 5 containers deep and just delete the grenade. */
				s = 0; 
 
				for (tobj = i; tobj; tobj = tobj->in_obj) {
					if (IN_ROOM(tobj) != NOWHERE) { 
						t = IN_ROOM(tobj);
						break;
					} else {
						if ((tch = CARRIED_BY(tobj))) { 
							t = IN_ROOM(tch);
							break;
						} else  {
							if ((tch = WORN_BY(tobj))) { 
								t = IN_ROOM(tch);
								break;
							}
						}
					}
					
					if (s++ >= 5)
						break;
				}
 
				/* then truly this grenade is nowhere?!? */
				if (t == NOWHERE) {
					log("SYSERR: Grenade is NOWHERE!!\r\n");
					extract_obj(i);
				} else { /* ok we have a room to blow up */

					/* peaceful rooms */
					if (ROOM_FLAGGED(t, ROOM_PEACEFUL)) {
						send_to_room(t, "You hear %s explode harmlessly, with a "
										 "loud POP!\n\r", i->short_description);
						extract_obj(i);
						return;
					}
 
					dam = dice(GET_OBJ_VAL(i, 1), GET_OBJ_VAL(i, 2));
 
					send_to_room(t, "Oh no - %s explodes!  KABOOOOOOOOOM!!!\r\n", 
									 i->short_description);
 
					for (door = 0; door < NUM_OF_DIRS; door++) {
						if (CAN_GO2(t, door)) 
							send_to_room(world[t].dir_option[door]->to_room,
											 "You hear a loud explosion!\r\n");
											 
					}
					
					for (tch = world[t].people; tch; tch = next_tch) {
						next_tch = tch->next_in_room;
 
						if (GET_POS(tch) <= POS_DEAD) {
							log("SYSERR: Attempt to damage corpse '%s' in room "
								 "#%d by '%s'.", GET_NAME(tch), 
								 GET_ROOM_VNUM(IN_ROOM(tch)), i->short_description);
							continue;
						}
 
						/* You can't damage an immortal! */
						if (IS_NPC(tch) || 
							 (pk_allowed && GET_LEVEL(tch) < LVL_IMMORT)) {
							GET_HIT(tch) -= dam;
							act("$n is caught in the blast!", 
								 FALSE, tch, 0, 0, TO_ROOM);
							act("You are caught in the blast!",
								 TRUE, tch, 0, 0, TO_CHAR);
							update_pos(tch);
 
							if (GET_POS(tch) <= POS_DEAD)
								die(tch, NULL);
						}
					} /* for (tch = world[t].people; tch; tch = next_tch) { */

					/* Ok hit all the people, now get rid of the grenade
					 * and any container it might have been in */
					extract_obj(i);
				} /* if (t == NOWHERE) { ... } else { */ 
			} /* if (GET_OBJ_VAL(i, 0) > 0) ... else { */
		} /* if (GET_OBJ_EXTRA(i).ITEM_LIVE_GRENADE == 1) { */
   } /* for (i = object_list; i; i = i->next) { */
}
 
